import asyncio
import logging
import time
from collections import OrderedDict
from datetime import datetime
from enum import Enum
from typing import (
    Mapping,
    Optional,
    List,
    Tuple,
    Union,
    Sequence,
    AsyncGenerator,
    Iterable,
    TYPE_CHECKING,
)

from aioredis import ConnectionClosedError, ReplyError
from aioredis.commands import streams
from aioredis.util import decode

from lightbus.client.utilities import queue_exception_checker, ErrorQueueType
from lightbus.transports.base import EventTransport, EventMessage
from lightbus.log import LBullets, L, Bold
from lightbus.serializers import ByFieldMessageSerializer, ByFieldMessageDeserializer
from lightbus.transports.redis.utilities import (
    RedisEventMessage,
    RedisTransportMixin,
    normalise_since_value,
    datetime_to_redis_steam_id,
    redis_stream_id_add_one,
    redis_stream_id_subtract_one,
    retry_on_redis_connection_failure,
)
from lightbus.utilities.async_tools import cancel
from lightbus.utilities.internal_queue import InternalQueue
from lightbus.utilities.frozendict import frozendict
from lightbus.utilities.human import human_time
from lightbus.utilities.importing import import_from_string

if TYPE_CHECKING:
    # pylint: disable=unused-import,cyclic-import
    from lightbus.config import Config
    from lightbus.client import BusClient


logger = logging.getLogger("lightbus.transports.redis")

Since = Union[str, datetime, None]


class StreamUse(Enum):
    PER_API = "per_api"
    PER_EVENT = "per_event"

    def __eq__(self, other):
        # pylint: disable=comparison-with-callable
        if isinstance(other, str):
            return self.value == other
        else:
            return super().__eq__(other)

    # Need to define this manually because we also
    # defined __eq__
    __hash__ = Enum.__hash__


class RedisEventTransport(RedisTransportMixin, EventTransport):
    """Redis Event Transport

    For a description of the protocol see https://lightbus.org/reference/protocols/event/
    """

    def __init__(
        self,
        redis_pool=None,
        *,
        service_name: str,
        consumer_name: str,
        url=None,
        serializer=ByFieldMessageSerializer(),
        deserializer=ByFieldMessageDeserializer(RedisEventMessage),
        connection_parameters: Mapping = frozendict(maxsize=100),
        batch_size=10,
        reclaim_batch_size: int = None,
        acknowledgement_timeout: float = 60,
        reclaim_interval: float = 60,
        max_stream_length: Optional[int] = 100_000,
        stream_use: StreamUse = StreamUse.PER_API,
        consumption_restart_delay: int = 5,
        consumer_ttl: int = 2_592_000,
    ):
        self.set_redis_pool(redis_pool, url, connection_parameters)
        self.batch_size = batch_size
        self.reclaim_batch_size = reclaim_batch_size if reclaim_batch_size else batch_size * 10
        self.service_name = service_name
        self.consumer_name = consumer_name
        self.acknowledgement_timeout = acknowledgement_timeout
        self.reclaim_interval = reclaim_interval
        self.max_stream_length = max_stream_length
        self.stream_use = stream_use
        self.consumption_restart_delay = consumption_restart_delay
        self.consumer_ttl = consumer_ttl
        super().__init__(serializer=serializer, deserializer=deserializer)

    @classmethod
    def from_config(
        cls,
        config: "Config",
        service_name: str = None,
        consumer_name: str = None,
        url: str = "redis://127.0.0.1:6379/0",
        connection_parameters: Mapping = frozendict(maxsize=100),
        batch_size: int = 10,
        reclaim_batch_size: int = None,
        serializer: str = "lightbus.serializers.ByFieldMessageSerializer",
        deserializer: str = "lightbus.serializers.ByFieldMessageDeserializer",
        acknowledgement_timeout: float = 60,
        reclaim_interval: Optional[float] = None,
        max_stream_length: Optional[int] = 100_000,
        stream_use: StreamUse = StreamUse.PER_API,
        consumption_restart_delay: int = 5,
        consumer_ttl: int = 2_592_000,
    ):
        serializer = import_from_string(serializer)()
        deserializer = import_from_string(deserializer)(RedisEventMessage)
        service_name = service_name or config.service_name
        consumer_name = consumer_name or config.process_name
        if isinstance(stream_use, str):
            stream_use = StreamUse[stream_use.upper()]

        if reclaim_interval is None:
            reclaim_interval = acknowledgement_timeout

        return cls(
            redis_pool=None,
            service_name=service_name,
            consumer_name=consumer_name,
            url=url,
            connection_parameters=connection_parameters,
            batch_size=batch_size,
            reclaim_batch_size=reclaim_batch_size,
            serializer=serializer,
            deserializer=deserializer,
            acknowledgement_timeout=acknowledgement_timeout,
            reclaim_interval=reclaim_interval,
            max_stream_length=max_stream_length or None,
            stream_use=stream_use,
            consumption_restart_delay=consumption_restart_delay,
            consumer_ttl=consumer_ttl,
        )

    async def send_event(self, event_message: EventMessage, options: dict):
        """Publish an event"""
        stream = self._get_stream_names(
            listen_for=[(event_message.api_name, event_message.event_name)]
        )[0]

        logger.debug(
            LBullets(
                L(
                    "Enqueuing event message {} in Redis stream {}",
                    Bold(event_message),
                    Bold(stream),
                ),
                items=dict(**event_message.get_metadata(), kwargs=event_message.get_kwargs()),
            )
        )

        # Performance: I suspect getting a connection from the connection manager each time is causing
        # performance issues. Need to confirm.
        with await self.connection_manager() as redis:
            start_time = time.time()
            await redis.xadd(
                stream=stream,
                fields=self.serializer(event_message),
                max_len=self.max_stream_length or None,
                exact_len=False,
            )

        logger.debug(
            L(
                "Enqueued event message {} in Redis in {} stream {}",
                Bold(event_message.canonical_name),
                human_time(time.time() - start_time),
                Bold(stream),
            )
        )

    async def consume(
        self,
        listen_for: List[Tuple[str, str]],
        listener_name: str,
        error_queue: ErrorQueueType,
        since: Union[Since, Sequence[Since]] = "$",
        forever=True,
    ) -> AsyncGenerator[List[RedisEventMessage], None]:
        """Consume events for the given APIs"""
        self._sanity_check_listen_for(listen_for)

        consumer_group = f"{self.service_name}-{listener_name}"

        if not isinstance(since, (list, tuple)):
            # Since has been specified as a single value. Normalise it into
            # the value-per-listener format.
            since = [since] * len(listen_for)
        since = map(normalise_since_value, since)

        stream_names = self._get_stream_names(listen_for)
        # Keys are stream names, values as the latest ID consumed from that stream
        streams = OrderedDict(zip(stream_names, since))
        expected_events = {event_name for _, event_name in listen_for}

        logger.debug(
            LBullets(
                L(
                    "Consuming events as consumer {} in group {} on streams",
                    Bold(self.consumer_name),
                    Bold(consumer_group),
                ),
                items={"{} ({})".format(*v) for v in streams.items()},
            )
        )

        # Cleanup any old groups & consumers
        await self._cleanup(stream_names)

        # Here we use a queue to combine messages coming from both the
        # fetch messages loop and the reclaim messages loop.
        queue = InternalQueue(maxsize=1)
        initial_reclaiming_complete = asyncio.Event()

        async def consume_loop():
            """Regular event consuming. See _fetch_new_messages()"""
            logger.debug(
                "Will begin consuming events once the initial event reclaiming is complete"
            )
            await initial_reclaiming_complete.wait()
            logger.debug("Event reclaiming is complete, beginning to consume events")

            async for messages in self._fetch_new_messages(
                streams, consumer_group, expected_events, forever
            ):
                await queue.put(messages)
                # Wait for the queue to empty before getting trying to get another message
                await queue.join()

        retry_consume_loop = retry_on_redis_connection_failure(
            fn=consume_loop, retry_delay=self.consumption_restart_delay, action="consuming events"
        )

        async def reclaim_loop():
            """
            Reclaim messages which other consumers have failed to
            processes in reasonable time. See _reclaim_lost_messages()
            """
            while True:
                logger.debug("Checking for any events which need reclaiming")
                async for messages in self._reclaim_lost_messages(
                    stream_names, consumer_group, expected_events
                ):
                    await queue.put(messages)
                    # Wait for the queue to empty before getting trying to get another message
                    await queue.join()

                initial_reclaiming_complete.set()
                await asyncio.sleep(self.reclaim_interval)

        consume_task = None
        reclaim_task = None

        try:
            # Run the two above coroutines in their own tasks
            consume_task = asyncio.ensure_future(
                queue_exception_checker(retry_consume_loop, error_queue)
            )
            reclaim_task = asyncio.ensure_future(
                queue_exception_checker(reclaim_loop(), error_queue)
            )

            while True:
                try:
                    messages = await queue.get()
                    logger.debug(
                        f"Got batch of {len(messages)} message(s). Yielding messages to Lightbus"
                        " client"
                    )
                    yield messages
                    logger.debug(
                        f"Batch of {len(messages)} message(s) was processed by Lightbus client."
                        " Marking as done."
                    )
                    queue.task_done()
                except GeneratorExit:
                    return
        finally:
            # Make sure we cleanup the tasks we created
            await cancel(consume_task, reclaim_task)

    async def _fetch_new_messages(
        self, streams, consumer_group, expected_events, forever
    ) -> AsyncGenerator[List[EventMessage], None]:
        """Coroutine to consume new messages

        The consumption has two stages:

          1. Fetch and yield any messages this consumer is responsible for processing but has yet
             to successfully process. This can happen in cases where a message was
             previously consumed but not acknowledged (i.e. due to an error).
             This is a one-off startup stage.
          2. Wait for new messages to arrive. Yield these messages when they arrive, then
             resume waiting for messages

        See Also:

            _reclaim_lost_messages() - Another coroutine which reclaims messages which timed out
                                       while being processed by other consumers in this group

        """
        with await self.connection_manager() as redis:
            # Firstly create the consumer group if we need to
            await self._create_consumer_groups(streams, redis, consumer_group)

            # Get any messages that this consumer has yet to process.
            # This can happen in the case where the processes died before acknowledging.
            pending_messages = await redis.xread_group(
                group_name=consumer_group,
                consumer_name=self.consumer_name,
                streams=list(streams.keys()),
                # Using ID '0' indicates we want unacked pending messages
                latest_ids=["0"] * len(streams),
                timeout=None,  # Don't block, return immediately
            )

            event_messages = []
            for stream, message_id, fields in pending_messages:
                message_id = decode(message_id, "utf8")
                stream = decode(stream, "utf8")
                try:
                    event_message = self._fields_to_message(
                        fields,
                        expected_events,
                        stream=stream,
                        native_id=message_id,
                        consumer_group=consumer_group,
                    )
                except (NoopMessage, IgnoreMessage):
                    # This listener doesn't need to care about this message, so acknowledge
                    # it and move on with our lives
                    await redis.xack(stream, consumer_group, message_id)
                    continue

                logger.debug(
                    LBullets(
                        L(
                            "⬅ Receiving pending event {} on stream {}",
                            Bold(message_id),
                            Bold(stream),
                        ),
                        items=dict(
                            **event_message.get_metadata(), kwargs=event_message.get_kwargs()
                        ),
                    )
                )
                event_messages.append(event_message)

            if event_messages:
                yield event_messages

            # We've now cleaned up any old messages that were hanging around.
            # Now we get on to the main loop which blocks and waits for new messages

            while True:
                # Fetch some messages.
                # This will block until there are some messages available
                try:
                    stream_messages = await redis.xread_group(
                        group_name=consumer_group,
                        consumer_name=self.consumer_name,
                        streams=list(streams.keys()),
                        # Using ID '>' indicates we only want new messages which have not
                        # been passed to other consumers in this group
                        latest_ids=[">"] * len(streams),
                        count=self.batch_size,
                    )
                except asyncio.CancelledError:
                    # We need to manually close the connection here otherwise the aioredis
                    # pool will emit warnings saying that this connection still has pending
                    # commands (i.e. the above blocking pop)
                    redis.close()
                    raise

                # Handle the messages we have received
                event_messages = []
                for stream, message_id, fields in stream_messages:
                    message_id = decode(message_id, "utf8")
                    stream = decode(stream, "utf8")
                    try:
                        event_message = self._fields_to_message(
                            fields,
                            expected_events,
                            stream=stream,
                            native_id=message_id,
                            consumer_group=consumer_group,
                        )
                    except (NoopMessage, IgnoreMessage):
                        # This listener doesn't need to care about this message, so acknowledge
                        # it and move on with our lives
                        logger.debug(
                            f"Ignoring NOOP event with ID {message_id} discovered during streaming"
                            " of messages"
                        )
                        await redis.xack(stream, consumer_group, message_id)
                        continue

                    logger.debug(
                        LBullets(
                            L(
                                "⬅ Received new event {} on stream {}",
                                Bold(message_id),
                                Bold(stream),
                            ),
                            items=dict(
                                **event_message.get_metadata(), kwargs=event_message.get_kwargs()
                            ),
                        )
                    )

                    event_messages.append(event_message)

                if event_messages:
                    yield event_messages

                if not forever:
                    return

    async def _reclaim_lost_messages(
        self, stream_names: List[str], consumer_group: str, expected_events: set
    ) -> AsyncGenerator[List[EventMessage], None]:
        """Reclaim batches of messages that other consumers in the group failed to acknowledge within a timeout.

        The timeout period is specified by the `acknowledgement_timeout` option.
        """
        with await self.connection_manager() as redis:
            for stream in stream_names:

                old_messages = True
                reclaim_from = None

                # Keep pulling reclaimable messages from Redis until there are none left
                while old_messages:
                    # reclaim_from keeps track of where we are up to in our fetching
                    # of messages
                    if not reclaim_from:
                        # This is our first iteration, so fetch from the start of time
                        reclaim_from = "-"
                    else:
                        # This is a subsequent iteration. XPENDING's 'start' parameter is inclusive,
                        # so we need to add one to the reclaim_from value to ensure we don't get a message
                        # we've already seen
                        reclaim_from = redis_stream_id_add_one(reclaim_from)

                    # Fetch the next batch of messages
                    try:
                        old_messages = await redis.xpending(
                            stream, consumer_group, reclaim_from, "+", count=self.reclaim_batch_size
                        )
                    except ReplyError as e:
                        if "NOGROUP" in str(e):
                            # Group or consumer doesn't exist yet, so stop processing for this loop.
                            break
                        else:
                            raise

                    timeout = self.acknowledgement_timeout * 1000
                    event_messages = []

                    # Try to claim each messages
                    for (
                        message_id,
                        consumer_name,
                        ms_since_last_delivery,
                        num_deliveries,
                    ) in old_messages:
                        message_id = decode(message_id, "utf8")
                        consumer_name = decode(consumer_name, "utf8")
                        reclaim_from = message_id

                        # This 'if' is not strictly required as the subsequent call to xclaim
                        # will honor the timeout parameter. However, using this if here allows
                        # for more sane logging from the point of view of the user. Without it
                        # we would report that we were trying to claim messages which were
                        # clearly not timed out yet.
                        if ms_since_last_delivery > timeout:
                            logger.info(
                                L(
                                    "Found timed out event {} in stream {}. Abandoned by {}."
                                    " Attempting to reclaim...",
                                    Bold(message_id),
                                    Bold(stream),
                                    Bold(consumer_name),
                                )
                            )

                            # *Try* to claim the message...
                            result = await redis.xclaim(
                                stream, consumer_group, self.consumer_name, int(timeout), message_id
                            )

                            # Parse each message we managed to claim
                            for _, fields in result:
                                # Note that sometimes we will claim a message and it will be 'nil'.
                                # In this case the result will be (None, {}). We therefore do not
                                # rely on the values we get back from XCLAIM.
                                # I suspect this may happen if a stream has been trimmed, thereby causing
                                # un-acknowledged messages to be deleted from Redis.
                                try:
                                    event_message = self._fields_to_message(
                                        fields,
                                        expected_events,
                                        stream=stream,
                                        native_id=message_id,
                                        consumer_group=consumer_group,
                                    )
                                except (NoopMessage, IgnoreMessage):
                                    # This listener doesn't need to care about this message, so acknowledge
                                    # it and move on with our lives
                                    logger.debug(
                                        f"Ignoring NOOP event with ID {message_id} discovered"
                                        " during event reclaiming"
                                    )
                                    await redis.xack(stream, consumer_group, message_id)
                                    continue

                                logger.debug(
                                    LBullets(
                                        L(
                                            "⬅ Reclaimed timed out event {} on stream {}. Abandoned"
                                            " by {}.",
                                            Bold(message_id),
                                            Bold(stream),
                                            Bold(consumer_name),
                                        ),
                                        items=dict(
                                            **event_message.get_metadata(),
                                            kwargs=event_message.get_kwargs(),
                                        ),
                                    )
                                )
                                event_messages.append(event_message)

                    # And yield our batch of messages
                    if event_messages:
                        yield event_messages

    async def acknowledge(self, *event_messages: RedisEventMessage):
        """Acknowledge that a message has been successfully processed"""
        with await self.connection_manager() as redis:
            p = redis.pipeline()
            for event_message in event_messages:
                p.xack(event_message.stream, event_message.consumer_group, event_message.native_id)
                logging.debug(
                    f"Preparing to acknowledge message {event_message.id} (Native ID:"
                    f" {event_message.native_id})"
                )

            logger.debug(
                f"Batch acknowledging successful processing of {len(event_messages)} message."
            )
            await p.execute()

    async def history(
        self,
        api_name,
        event_name,
        start: datetime = None,
        stop: datetime = None,
        start_inclusive: bool = True,
        batch_size: int = 100,
    ) -> AsyncGenerator[EventMessage, None]:
        """Retrieve historical events for the given API

        Will not have any impact on existing consumer groups.
        """
        redis_start = datetime_to_redis_steam_id(start) if start else "-"
        redis_stop = datetime_to_redis_steam_id(stop) if stop else "+"

        if start and not start_inclusive:
            redis_start = redis_stream_id_add_one(redis_start)

        stream_name = self._get_stream_names([(api_name, event_name)])[0]

        logger.debug(
            f"Getting history for stream {stream_name} from {redis_start} ({start}) "
            f"to {redis_stop} ({stop}) in batches of {batch_size}"
        )

        with await self.connection_manager() as redis:
            messages = True
            while messages:
                messages = await redis.xrevrange(
                    stream_name, redis_stop, redis_start, count=batch_size
                )
                if not messages:
                    return
                for message_id, fields in messages:
                    message_id = decode(message_id, "utf8")
                    redis_stop = redis_stream_id_subtract_one(message_id)
                    try:
                        event_message = self._fields_to_message(
                            fields,
                            expected_event_names={event_name},
                            stream=stream_name,
                            native_id=message_id,
                            consumer_group=None,
                        )
                    except (NoopMessage, IgnoreMessage):
                        logger.debug(
                            f"Ignoring NOOP event with ID {message_id} discovered during fetching"
                            " of event history"
                        )
                    else:
                        yield event_message

    async def lag(
        self,
        api_name: str,
        event_name: str,
        listener_name: str,
        include_pending: bool = True,
        max_count: Optional[int] = 100,
    ) -> int:
        """How far behind is the given listener for the given api/event?

        This must fetch all lagging messages from Redis, which means
        this cannot be particularly performant for determining the size of large
        lags. The ideal is therefore to set max_count to a sane minimum, thereby
        minimising the number of messages which need to be pulled back.
        Note however, that this method will never return a value larger than
        max_count + the total pending messages.

        If max_count is None then the full lag will be calculated.
        """
        consumer_group = f"{self.service_name}-{listener_name}".encode("utf8")
        stream_name = self._get_stream_names([(api_name, event_name)])[0]

        with await self.connection_manager() as redis:
            try:
                # Get the groups for the given stream
                groups = await redis.xinfo_groups(stream_name)
            except ReplyError as e:
                if "no such key" in str(e):
                    # Steam doesn't exist, so group doesn't exist
                    groups = []
                else:
                    raise

            for group in groups:
                # Find the group we care about
                if group[b"name"] == consumer_group:
                    if max_count is not None:
                        # We should normally receive the last delivered message in addition to any lagging
                        # messages We take that into account here and is tested by test_lag_max_count_set()
                        max_count += 1

                    # This is the part with poor performance. Get all the lagging messages.
                    messages = await redis.xrange(
                        stream=stream_name, start=group[b"last-delivered-id"], count=max_count
                    )

                    # Messages found, exit our for loop
                    break
            else:
                # For loop didn't find any matching consumer groups, return 0
                return 0

        first_message_id = messages[0][0] if messages else None
        total_messages = len(messages)

        if first_message_id == group[b"last-delivered-id"]:
            # The last delivered message is among the results returned, so
            # take that off our count
            lag = total_messages - 1
        elif max_count is not None and total_messages == max_count:
            # The last delivered message has been deleted, and we received the max_count number of results,
            # which is actually one more than we expected because we incremented max_count above.
            # Therefore fix this by subtracting one
            lag = total_messages - 1
        else:
            # The last delivered message has been deleted,
            # so all returned messages are lagging. We retrieved all messages
            # (rather than limiting with max_count) therefore no need to
            # subtract anything
            lag = total_messages

        if include_pending:
            lag += group[b"pending"]

        return lag

    async def _create_consumer_groups(self, streams, redis, consumer_group):
        """Ensure the consumer groups exist

        This is means we have to ensure the streams exist too
        """
        for stream, since in streams.items():
            if not await redis.exists(stream):
                # Add a noop to ensure the stream exists
                # TODO: We can now use MKSTREAM, change this logic
                #       Documented here: https://redis.io/topics/streams-intro
                await redis.xadd(stream, fields={"": ""})

            try:
                # Create the group (it may already exist)
                await redis.xgroup_create(stream, consumer_group, latest_id=since)
            except ReplyError as e:
                if "BUSYGROUP" in str(e):
                    # Already exists
                    pass
                else:
                    raise

    async def _cleanup(self, stream_names: List[str]):
        """Cleanup old consumers and groups

        A group will be deleted if it contains no consumers.

        A consumer will be deleted if it has been idle for more than consumer_ttl.
        """
        if not self.consumer_ttl:
            # Don't do the cleanup if no TTL is given, consider this to mean
            # cleanup is disabled
            return

        with await self.connection_manager() as redis:
            # For every stream key...
            for stream_name in stream_names:
                consumers: List[Tuple[str, str]] = []

                # Get all the groups for that key...
                try:
                    groups = await redis.xinfo_groups(stream_name)
                except ReplyError as e:
                    if "ERR no such key" in str(e):
                        # Steam doesn't exist yet
                        groups = []
                    else:
                        raise

                for group in groups:
                    active_consumers = 0
                    group_name = group[b"name"]

                    try:
                        consumers = await redis.xinfo_consumers(stream_name, group_name)
                    except ReplyError as e:
                        # Group has already vanished,
                        # perhaps something else is doing a cleanup
                        if "NOGROUP" in str(e):
                            continue

                    # Get all the consumers for that group
                    for consumer in consumers:
                        consumer_name = consumer[b"name"]
                        idle_seconds = consumer[b"idle"] / 1000

                        # And delete the consumer if they have not re-started
                        # listening for self.consumer_ttl seconds
                        if idle_seconds >= self.consumer_ttl:
                            logger.debug(
                                f"Cleaning up consumer {consumer_name} in group {group_name} on"
                                f" stream {stream_name}. The consumer has been idle for"
                                f" {idle_seconds} seconds, which is more than the consumer TTL of"
                                f" {self.consumer_ttl}"
                            )
                            await redis.xgroup_delconsumer(stream_name, group_name, consumer_name)
                        else:
                            active_consumers += 1

                    # If no active consumers were found for this group, then delete the entire group
                    # on the grounds that it is no longer used and can be cleaned up.
                    if not active_consumers:
                        # We do this atomically using a lua script. This avoids race conditions
                        # whereby a new consumer comes into existence the moment before we delete the group
                        try:
                            await redis.eval(
                                ATOMIC_DESTROY_CONSUMER_GROUP, [stream_name], [group_name]
                            )
                        except ReplyError as e:
                            if "NOGROUP" in str(e):
                                # Already deleted
                                pass

    def _fields_to_message(
        self,
        fields: dict,
        expected_event_names: Iterable[str],
        stream: str,
        native_id: str,
        consumer_group: Optional[str],
    ) -> Optional[RedisEventMessage]:
        """Convert a dict of Redis message fields into a RedisEventMessage"""

        if not fields:
            # Redis returned us a message of 'nil'
            raise NoopMessage()

        if tuple(fields.items()) == ((b"", b""),):
            # Is a noop message, ignore
            raise NoopMessage()

        message = self.deserializer(
            fields, stream=stream, native_id=native_id, consumer_group=consumer_group
        )

        want_message = ("*" in expected_event_names) or (message.event_name in expected_event_names)
        if self.stream_use == StreamUse.PER_API and not want_message:
            # Only care about events we are listening for. If we have one stream
            # per API then we're probably going to receive some events we don't care about.
            logger.debug(
                f"Ignoring message for unneeded event: {message}. "
                f"Only listening for {', '.join(expected_event_names)}"
            )
            raise IgnoreMessage()
        return message

    def _get_stream_names(self, listen_for):
        """Convert a list of api names & event names into stream names

        The format of these names will vary based on the stream_use setting.
        """
        stream_names = []
        for api_name, event_name in listen_for:
            if self.stream_use == StreamUse.PER_EVENT:
                stream_name = f"{api_name}.{event_name}:stream"
            elif self.stream_use == StreamUse.PER_API:
                stream_name = f"{api_name}.*:stream"
            else:
                raise ValueError(
                    "Invalid value for stream_use config option. This should have been caught "
                    "during config validation."
                )
            if stream_name not in stream_names:
                stream_names.append(stream_name)
        return stream_names


class NoopMessage(Exception):
    pass


class IgnoreMessage(Exception):
    pass


# See RedisEventTransport._cleanup()
ATOMIC_DESTROY_CONSUMER_GROUP = """
local stream_name = KEYS[1]
local group_name = ARGV[1]
local consumers = redis.call('xinfo', 'consumers', stream_name, group_name)
if table.getn(consumers) == 0 then
    redis.call('xgroup', 'destroy', stream_name, group_name)
end
"""


def parse_messages(messages):
    # TODO: Remove and bump required aioredis version when merged:
    #       https://github.com/aio-libs/aioredis/pull/723
    if messages is None:
        return []

    parsed_messages = []
    for message in messages:
        if message is None:
            # In some conditions redis will return a NIL message
            parsed_messages.append((None, {}))
        else:
            mid, values = message
            parsed_messages.append((mid, streams.fields_to_dict(values)))

    return parsed_messages


streams.parse_messages = parse_messages



import inspect
import json
import logging
from json import JSONDecodeError
from pathlib import Path
from typing import Optional, TextIO, Union, ChainMap, List, Tuple, Dict, TYPE_CHECKING
import asyncio
import itertools
import sys

import jsonschema

from lightbus.exceptions import (
    InvalidApiForSchemaCreation,
    InvalidSchema,
    SchemaNotFound,
    ValidationError,
    RemoteSchemasNotLoaded,
    RemoteSchemasNotLoaded,
)
from lightbus.schema.encoder import json_encode
from lightbus.schema.hints_to_schema import (
    make_response_schema,
    make_rpc_parameter_schema,
    make_event_parameter_schema,
)
from lightbus.transports.registry import SchemaTransportPoolType
from lightbus.utilities.io import make_file_safe_api_name
from lightbus.api import Api, Event

if TYPE_CHECKING:
    # pylint: disable=unused-import,cyclic-import
    from lightbus.transports.base import SchemaTransport
    from lightbus.transports.pool import TransportPool

logger = logging.getLogger(__name__)


class Schema:
    """ Represents the bus' schema


    Note that the presence of a schema does not necessarily
    indicate that a lightbus process is present or ready to serve
    requests for the API. For that you will need to consume the events
    produced by the state plugin.

    That being said, you should expect old schemas to be dropped
    after max_age_seconds.

    """

    def __init__(
        self,
        schema_transport: "SchemaTransportPoolType",
        max_age_seconds: Optional[int] = 60,
        human_readable: bool = True,
    ):
        self.schema_transport = schema_transport
        self._schema_transport: Optional["SchemaTransport"] = None
        self.max_age_seconds = max_age_seconds
        self.human_readable = human_readable

        # Schemas which have been provided locally. These will either be locally-available
        # APIs, or schemas which have been loaded from local files
        self.local_schemas = {}

        # Schemas which have been retrieved from the bus. This will also contain local
        # schemas which have been stored onto the bus. The storing and retrieving of
        # remote schemas is mediated by the schema transport.
        self._remote_schemas: Optional[Dict[str, dict]] = None

    def __contains__(self, item):
        return item in self.local_schemas or item in self.remote_schemas

    async def add_api(self, api: "Api"):
        """Adds an API locally, and sends to the transport"""
        schema = api_to_schema(api)
        self.local_schemas[api.meta.name] = schema
        await self.schema_transport.store(api.meta.name, schema, ttl_seconds=self.max_age_seconds)

    def get_api_schema(self, api_name) -> Optional[dict]:
        """Get the schema for the given API"""
        api_schema = self.local_schemas.get(api_name) or self.remote_schemas.get(api_name)
        if not api_schema:
            # TODO: Add link to docs in error message
            raise SchemaNotFound(
                "No schema could be found for API {}. You should ensure that either this "
                "API is being served by another lightbus process, or you can load this schema manually."
                "".format(api_name)
            )
        return api_schema

    def get_event_schema(self, api_name, event_name):
        event_schemas = self.get_api_schema(api_name)["events"]
        try:
            return event_schemas[event_name]
        except KeyError:
            raise SchemaNotFound(
                "Found schema for API '{}', but it did not contain an event named '{}'"
                "".format(api_name, event_name)
            )

    def get_rpc_schema(self, api_name, rpc_name):
        rpc_schemas = self.get_api_schema(api_name)["rpcs"]
        try:
            return rpc_schemas[rpc_name]
        except KeyError:
            raise SchemaNotFound(
                "Found schema for API '{}', but it did not contain a RPC named '{}'"
                "".format(api_name, rpc_name)
            )

    def get_event_or_rpc_schema(self, api_name, name):
        try:
            return self.get_event_schema(api_name, name)
        except SchemaNotFound:
            pass

        try:
            return self.get_rpc_schema(api_name, name)
        except SchemaNotFound:
            pass

        # TODO: Add link to docs in error message
        raise SchemaNotFound(
            "No schema found for '{}' on API '{}'. You should either, a) ensure this "
            "API is being served by another lightbus process, or b) load this schema manually."
            "".format(name, api_name)
        )

    def validate_parameters(self, api_name, event_or_rpc_name, parameters):
        """Validate the parameters for the given event/rpc

        This will raise an `jsonschema.ValidationError` exception on error,
        or return None if valid.
        """
        json_schema = self.get_event_or_rpc_schema(api_name, event_or_rpc_name)["parameters"]
        try:
            jsonschema.validate(parameters, json_schema)
        except jsonschema.ValidationError as e:
            logger.error(e)
            path = list(e.absolute_path)
            if not path:
                raise ValidationError(
                    f"Validation error when using JSON schema to validate parameters for \n"
                    f"{api_name}.{event_or_rpc_name}.\n"
                    f"\n"
                    f"It is likely you have included an unwanted parameter or omitted a required \n"
                    f"parameter.\n"
                    f"\n"
                    f"The error was: {e.message}\n"
                    f"\n"
                    f"The full validator error was logged above"
                ) from None
            elif len(path) == 1:
                raise ValidationError(
                    f"Validation error when using JSON schema to validate parameters for \n"
                    f"{api_name}.{event_or_rpc_name}.\n"
                    f"\n"
                    f"It is likely that you have passed in an invalid value for the \n"
                    f"'{path[0]}' parameter.\n"
                    f"\n"
                    f"The error given was: {e.message}\n"
                    f"\n"
                    f"The full validator error was logged above"
                ) from None
            else:
                raise ValidationError(
                    f"Validation error when using JSON schema to validate parameters for \n"
                    f"{api_name}.{event_or_rpc_name}.\n"
                    f"\n"
                    f"This was an error in validating the internal structure of one \n"
                    f"of the parameters' values. The path to this error is \n"
                    f"'<root>.{'.'.join(e.absolute_path)}'.\n"
                    f"\n"
                    f"The error given was: {e.message}\n"
                    f"\n"
                    f"The full validator error was logged above"
                ) from None

    def validate_response(self, api_name, rpc_name, response):
        """Validate the parameters for the given event/rpc

        This will raise an `jsonschema.ValidationError` exception on error,
        or return None if valid.

        Note that only RPCs have responses. Accessing this property for an
        event will result in a SchemaNotFound error.
        """
        json_schema = self.get_rpc_schema(api_name, rpc_name)["response"]
        try:
            jsonschema.validate(response, json_schema)
        except jsonschema.ValidationError as e:
            logger.error(e)
            path = list(e.absolute_path)
            if not path:
                raise ValidationError(
                    f"Validation error when using JSON schema to validate result from \n"
                    f"RPC {api_name}.{rpc_name}.\n"
                    f"\n"
                    f"It is likely the response was either of the incorrect type, or "
                    f"some fields were erroneously absent/present.\n"
                    f"\n"
                    f"The error was: {e.message}\n"
                    f"\n"
                    f"The full validator error was logged above"
                ) from None
            else:
                raise ValidationError(
                    f"Validation error when using JSON schema to validate result from \n"
                    f"RPC {api_name}.{rpc_name}.\n"
                    f"\n"
                    f"This was an error in validating the internal structure of the \n"
                    f"data returned values. The path to this error is \n"
                    f"'<root>.{'.'.join(e.absolute_path)}'.\n"
                    f"\n"
                    f"The error given was: {e.message}\n"
                    f"\n"
                    f"The full validator error was logged above"
                ) from None

    @property
    def api_names(self) -> List[str]:
        return list(set(itertools.chain(self.local_schemas.keys(), self.remote_schemas.keys())))

    @property
    def events(self) -> List[Tuple[str, str]]:
        """Get a list of all events available on the bus

        Each event is a tuple in the form `(api_name, event_name)`
        """
        events = []
        for api_name in self.api_names:
            api_schema = self.get_api_schema(api_name)
            if api_schema:
                for event_name in api_schema["events"].keys():
                    events.append((api_name, event_name))
        return events

    @property
    def rpcs(self) -> List[Tuple[str, str]]:
        """Get a list of all RPCs available on the bus

        Each rpc is a tuple in the form `(api_name, rpc_name)`
        """
        rpcs = []
        for api_name in self.api_names:
            api_schema = self.get_api_schema(api_name)
            if api_schema:
                for event_name in api_schema["rpcs"].keys():
                    rpcs.append((api_name, event_name))
        return rpcs

    async def save_to_bus(self):
        """Save the schema onto the bus

        This will be done using the `schema_transport` provided to `__init__()`
        """
        for api_name, schema in self.local_schemas.items():
            await self.schema_transport.store(api_name, schema, ttl_seconds=self.max_age_seconds)

    async def load_from_bus(self):
        """Save the schema from the bus

        This will be done using the `schema_transport` provided to `__init__()`
        """
        self._remote_schemas = await self.schema_transport.load()

    async def ensure_loaded_from_bus(self):
        if self._remote_schemas is None:
            await self.load_from_bus()

    @property
    def remote_schemas(self) -> Dict[str, Dict]:
        """Schemas which have been retrieved from the bus.

        This will also contain local schemas which have been stored onto the bus. \
        The storing and retrieving of remote schemas is mediated by the schema transport.

        The returned value is a dictionary where keys are fully qualified API names,
        and the values are JSON schemas
        """
        if self._remote_schemas is None:
            raise RemoteSchemasNotLoaded(
                "The remote schemas have not yet been loaded. Lightbus should have ensured this was done "
                "already, and therefore this is likely a bug. However, calling "
                "bus.client.lazy_load_now() should resolve this."
            )
        return self._remote_schemas

    async def monitor(self, interval=None):
        """Monitor for remote schema changes and keep any local schemas alive on the bus
        """
        interval = interval or self.max_age_seconds * 0.8
        try:
            while True:
                await asyncio.sleep(interval)
                # Keep alive our local schemas
                for api_name, schema in self.local_schemas.items():
                    await self.schema_transport.ping(
                        api_name, schema, ttl_seconds=self.max_age_seconds
                    )

                # Read the entire schema back from the bus
                await self.load_from_bus()
        except asyncio.CancelledError:
            return

    def save_local(self, destination: Union[str, Path, TextIO] = None):
        """Save all present schemas to a local file

        This will save both local & remote schemas to a local file
        """
        if isinstance(destination, str):
            destination = Path(destination)

        if destination is None:
            self._dump_to_file(sys.stdout)
            sys.stdout.write("\n")
        elif destination.is_dir():
            self._dump_to_directory(destination)
        else:
            with destination.open("w", encoding="utf8") as f:
                self._dump_to_file(f)

    def load_local(self, source: Union[str, Path, TextIO] = None):
        """Load schemas from a local file

        These files will be treated as local schemas, and will not be sent to the bus.
        This can be useful for validation during development and testing.
        """
        if isinstance(source, str):
            source = Path(source)

        def _load_schema(path, file_data):
            try:
                return json.loads(file_data)
            except JSONDecodeError as e:
                raise InvalidSchema("Could not parse schema file {}: {}".format(path, e.msg))

        if source is None:
            # No source, read from stdin
            schema = _load_schema("[stdin]", sys.stdin.read())
        elif hasattr(source, "is_dir") and source.is_dir():
            # Read each json file in directory
            schemas = []
            for file_path in source.glob("*.json"):
                schemas.append(_load_schema(file_path, file_path.read_text(encoding="utf8")))
            schema = ChainMap(*schemas)
        elif hasattr(source, "read"):
            # Read file handle
            schema = _load_schema(source.name, source.read())
        elif hasattr(source, "read_text"):
            # Read pathlib Path
            schema = _load_schema(source.name, source.read_text())
        else:
            raise InvalidSchema(
                "Did not recognise provided source as either a "
                "directory path, file path, or file handle: {}".format(source)
            )

        for api_name, api_schema in schema.items():
            self.local_schemas[api_name] = api_schema

        return schema

    def _dump_to_directory(self, destination: Path):
        for api_name in self.api_names:
            file_name = "{}.json".format(make_file_safe_api_name(api_name))
            (destination / file_name).write_text(self._get_dump(api_name), encoding="utf8")

    def _dump_to_file(self, f):
        f.write(self._get_dump())

    def _get_dump(self, api_name=None):
        if api_name:
            schema = {api_name: self.get_api_schema(api_name)}
        else:
            schema = {api_name: self.get_api_schema(api_name) for api_name in self.api_names}

        indent = 2 if self.human_readable else None
        return json_encode(schema, indent=indent)

    async def close(self):
        await self.schema_transport.close()


class Parameter(inspect.Parameter):
    """Describes the name and type of an event parameter"""

    empty = inspect.Parameter.empty

    def __init__(self, name, annotation=empty, *, default=empty):
        super(Parameter, self).__init__(
            name, inspect.Parameter.KEYWORD_ONLY, default=default, annotation=annotation
        )


class WildcardParameter(inspect.Parameter):
    """Describes a **kwargs style parameter to an event
    """

    def __init__(self):
        super(WildcardParameter, self).__init__(
            name="kwargs", kind=inspect.Parameter.VAR_KEYWORD, default={}, annotation=dict
        )


def api_to_schema(api: "lightbus.Api") -> dict:
    """Produce a lightbus schema for the given API"""
    schema = {"rpcs": {}, "events": {}}

    if isinstance(api, type):
        raise InvalidApiForSchemaCreation(
            "An attempt was made to derive an API schema from a type/class, rather than "
            "from an instance of an API. This is probably because you are passing an API "
            "class to api_to_schema(), rather than an instance of the API class."
        )

    for member_name, member in inspect.getmembers(api):
        if member_name.startswith("_"):
            # Don't create schema from private methods
            continue
        if hasattr(Api, member_name):
            # Don't create schema for methods defined on Api class
            continue

        if inspect.ismethod(member):
            schema["rpcs"][member_name] = {
                "parameters": make_rpc_parameter_schema(api.meta.name, member_name, method=member),
                "response": make_response_schema(api.meta.name, member_name, method=member),
            }
        elif isinstance(member, Event):
            schema["events"][member_name] = {
                "parameters": make_event_parameter_schema(api.meta.name, member_name, event=member)
            }

    return schema



from json import JSONEncoder


def json_encode(obj, indent=2, sort_keys=True, **options):
    # TODO: This is also used for non-schema related encoding. Either move
    #       this elsewhere, or create a new general purpose encoder
    return JSONEncoder(indent=indent, sort_keys=sort_keys, **options).encode(obj)



import logging
from typing import Union

from lightbus.exceptions import UnknownApi
from lightbus.message import EventMessage, RpcMessage, ResultMessage
from lightbus.schema import Schema
from lightbus.config import Config

logger = logging.getLogger(__name__)


def validate_incoming(
    config: Config, schema: Schema, message: Union[EventMessage, RpcMessage, ResultMessage]
):
    return _validate(config, schema, message, "incoming")


def validate_outgoing(
    config: Config, schema: Schema, message: Union[EventMessage, RpcMessage, ResultMessage]
):
    return _validate(config, schema, message, "outgoing")


def _validate(
    config, schema, message: Union[EventMessage, RpcMessage, ResultMessage], direction: str
):
    if direction not in ("incoming", "outgoing"):
        raise AssertionError("Invalid direction specified")

    # Result messages do not carry the api or procedure name, so allow them to be
    # specified manually
    api_name = message.api_name
    event_or_rpc_name = getattr(message, "procedure_name", None) or getattr(message, "event_name")
    api_config = config.api(api_name)
    strict_validation = api_config.strict_validation

    if not getattr(api_config.validate, direction):
        return

    if api_name not in schema:
        if strict_validation:
            raise UnknownApi(
                f"Validation is enabled for API named '{api_name}', but there is no schema present for this API. "
                f"Validation is therefore not possible. You are also seeing this error because the "
                f"'strict_validation' setting is enabled. Disabling this setting will turn this exception "
                f"into a warning. "
            )
        else:
            logger.warning(
                f"Validation is enabled for API named '{api_name}', but there is no schema present for this API. "
                f"Validation is therefore not possible. You can force this to be an error by enabling "
                f"the 'strict_validation' config option. You can silence this message by disabling validation "
                f"for this API using the 'validate' option."
            )
            return

    if isinstance(message, (RpcMessage, EventMessage)):
        schema.validate_parameters(api_name, event_or_rpc_name, message.kwargs)
    elif isinstance(message, ResultMessage):
        schema.validate_response(api_name, event_or_rpc_name, message.result)



import asyncio
import logging
import time
from asyncio import iscoroutinefunction, iscoroutine
from typing import List

from lightbus.api import Api
from lightbus.client import commands
from lightbus.client.commands import ConsumeRpcsCommand
from lightbus.client.subclients.base import BaseSubClient
from lightbus.client.utilities import validate_event_or_rpc_name, Error, ErrorQueueType
from lightbus.client.validator import validate_outgoing, validate_incoming
from lightbus.exceptions import (
    NoApisToListenOn,
    SuddenDeathException,
    LightbusWorkerError,
    LightbusTimeout,
)
from lightbus.log import L, Bold
from lightbus.message import ResultMessage, RpcMessage
from lightbus.utilities.async_tools import run_user_provided_callable, cancel
from lightbus.utilities.internal_queue import InternalQueue
from lightbus.utilities.casting import cast_to_signature
from lightbus.utilities.deforming import deform_to_bus
from lightbus.utilities.frozendict import frozendict
from lightbus.utilities.human import human_time
from lightbus.utilities.singledispatch import singledispatchmethod

logger = logging.getLogger(__name__)


async def bail_on_error(error_queue: ErrorQueueType, co):
    """Cancel the provided coroutine if an error appears on the error queue

    If no error appears on the queue, then the result of the
    coroutine will be returned.

    If an error does appear then the error will be raised
    """

    assert iscoroutine(co), "@bail_on_error only operates on coroutines"
    fn_task = asyncio.ensure_future(co)
    monitor_task = asyncio.ensure_future(error_queue.get())

    done, pending = await asyncio.wait({fn_task, monitor_task}, return_when=asyncio.FIRST_COMPLETED)

    if fn_task in done:
        # All ok
        await cancel(monitor_task)
        return fn_task.result()
    else:
        # An error appeared in the queue
        await cancel(fn_task)
        error: Error = monitor_task.result()
        # Put the error back on the queue so it can be picked up
        # by the bus client
        error_queue.put_nowait(error)
        raise error.value


class RpcResultClient(BaseSubClient):
    """Functionality for both RPCs and results"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    async def consume_rpcs(self, apis: List[Api] = None):
        """Start a background task to consume RPCs

        This will consumer RPCs on APIs which have been registered with this
        bus client.
        """
        if apis is None:
            apis = self.api_registry.all()

        if not apis:
            raise NoApisToListenOn(
                "No APIs to consume on in consume_rpcs(). Either this method was called with apis=[], "
                "or the API registry is empty."
            )

        api_names = [api.meta.name for api in apis]

        await self.producer.send(ConsumeRpcsCommand(api_names=api_names)).wait()

    async def call_rpc_remote(
        self, api_name: str, name: str, kwargs: dict = frozendict(), options: dict = frozendict()
    ):
        """ Perform an RPC call

        Call an RPC and return the result.
        """
        kwargs = deform_to_bus(kwargs)
        rpc_message = RpcMessage(api_name=api_name, procedure_name=name, kwargs=kwargs)
        validate_event_or_rpc_name(api_name, "rpc", name)

        logger.info("📞  Calling remote RPC {}.{}".format(Bold(api_name), Bold(name)))

        start_time = time.time()

        validate_outgoing(self.config, self.schema, rpc_message)

        await self.hook_registry.execute("before_rpc_call", rpc_message=rpc_message)

        result_queue = InternalQueue()

        # Send the RPC
        await self.producer.send(
            commands.CallRpcCommand(message=rpc_message, options=options)
        ).wait()

        # Start a listener which will wait for results
        await self.producer.send(
            commands.ReceiveResultCommand(
                message=rpc_message, destination_queue=result_queue, options=options
            )
        ).wait()

        # Wait for the result from the listener we started.
        # The RpcResultDock will handle timeouts
        result = await bail_on_error(self.error_queue, result_queue.get())

        call_time = time.time() - start_time

        try:
            if isinstance(result, Exception):
                raise result
        except asyncio.TimeoutError:
            raise LightbusTimeout(
                f"Timeout when calling RPC {rpc_message.canonical_name} after waiting for {human_time(call_time)}. "
                f"It is possible no Lightbus process is serving this API, or perhaps it is taking "
                f"too long to process the request. In which case consider raising the 'rpc_timeout' "
                f"config option."
            ) from None
        else:
            assert isinstance(result, ResultMessage)
            result_message = result

        await self.hook_registry.execute(
            "after_rpc_call", rpc_message=rpc_message, result_message=result_message
        )

        if not result_message.error:
            logger.info(
                L(
                    "🏁  Remote call of {} completed in {}",
                    Bold(rpc_message.canonical_name),
                    human_time(call_time),
                )
            )
        else:
            logger.warning(
                L(
                    "⚡ Error during remote call of RPC {}. Took {}: {}",
                    Bold(rpc_message.canonical_name),
                    human_time(call_time),
                    result_message.result,
                )
            )
            raise LightbusWorkerError(
                "Error while calling {}: {}\nRemote stack trace:\n{}".format(
                    rpc_message.canonical_name, result_message.result, result_message.trace
                )
            )

        validate_incoming(self.config, self.schema, result_message)

        return result_message.result

    async def _call_rpc_local(self, api_name: str, name: str, kwargs: dict = frozendict()):
        api = self.api_registry.get(api_name)
        validate_event_or_rpc_name(api_name, "rpc", name)

        start_time = time.time()
        try:
            method = getattr(api, name)
            if self.config.api(api_name).cast_values:
                kwargs = cast_to_signature(kwargs, method)
            result = await run_user_provided_callable(method, args=[], kwargs=kwargs)
        except (asyncio.CancelledError, SuddenDeathException):
            raise
        except Exception as e:
            logging.exception(e)
            logger.warning(
                L(
                    "⚡  Error while executing {}.{}. Took {}",
                    Bold(api_name),
                    Bold(name),
                    human_time(time.time() - start_time),
                )
            )
            raise
        else:
            logger.info(
                L(
                    "⚡  Executed {}.{} in {}",
                    Bold(api_name),
                    Bold(name),
                    human_time(time.time() - start_time),
                )
            )
            return result

    async def close(self):
        await super().close()
        await self.producer.send(commands.CloseCommand()).wait()

        await self.consumer.close()
        await self.producer.close()

    @singledispatchmethod
    async def handle(self, command):
        raise NotImplementedError(f"Did not recognise command {command.__class__.__name__}")

    @handle.register
    async def handle_execute_rpc(self, command: commands.ExecuteRpcCommand):
        await self.schema.ensure_loaded_from_bus()
        validate_incoming(self.config, self.schema, command.message)

        await self.hook_registry.execute("before_rpc_execution", rpc_message=command.message)
        try:
            result = await self._call_rpc_local(
                api_name=command.message.api_name,
                name=command.message.procedure_name,
                kwargs=command.message.kwargs,
            )
        except SuddenDeathException:
            # Used to simulate message failure for testing
            return
        except asyncio.CancelledError:
            raise
        except Exception as e:
            result = e
        else:
            result = deform_to_bus(result)

        result_message = ResultMessage(
            result=result,
            rpc_message_id=command.message.id,
            api_name=command.message.api_name,
            procedure_name=command.message.procedure_name,
        )
        await self.hook_registry.execute(
            "after_rpc_execution", rpc_message=command.message, result_message=result_message
        )

        if not result_message.error:
            validate_outgoing(self.config, self.schema, result_message)

        await self.producer.send(
            commands.SendResultCommand(message=result_message, rpc_message=command.message)
        ).wait()



import argparse
import logging
import sys

import lightbus
import lightbus.client
import lightbus.creation
from lightbus.config import Config
from lightbus.exceptions import FailedToImportBusModule
from lightbus.plugins import PluginRegistry
from lightbus.utilities.logging import configure_logging
from lightbus.utilities.async_tools import block
import lightbus.commands.run
import lightbus.commands.shell
import lightbus.commands.dump_schema
import lightbus.commands.dump_config_schema
import lightbus.commands.inspect
import lightbus.commands.version

logger = logging.getLogger(__name__)

COMMAND_PARSED_ARGS = {}

RED = "\033[91m" if sys.stdout.isatty() else ""
RESET = "\033[0m" if sys.stdout.isatty() else ""


def lightbus_entry_point():  # pragma: no cover
    sys.path.insert(0, "")
    configure_logging()
    run_command_from_args()


def run_command_from_args(args=None, **extra):
    parsed_args = parse_args(args)

    # Store the args we received away for later. The
    # bus creation process will use them as overrides.
    # This is a bit of hack to allow command line overrides
    # while having no control over bus instantiation (because
    # the application developer does that in bus.py)
    COMMAND_PARSED_ARGS.clear()
    COMMAND_PARSED_ARGS.update(dict(parsed_args._get_kwargs()))

    if hasattr(parsed_args, "config_file"):
        config = load_config(parsed_args)
        plugin_registry = PluginRegistry()
        plugin_registry.autoload_plugins(config)
    else:
        # Command didn't set up any of the common args (via setup_common_arguments()),
        # which are needed to load up the config. As the command didn't set these up,
        # then assume the command doesn't want to be passed the config & plugin_registry
        config = None
        plugin_registry = None

    try:
        if config is None:
            # Don't pass the config & plugin_registry, see above comment
            parsed_args.func(parsed_args, **extra)
        else:
            parsed_args.func(parsed_args, config, plugin_registry, **extra)

    except FailedToImportBusModule as e:
        sys.stderr.write(f"{RED}{e}{RESET}\n")


def parse_args(args=None):
    parser = argparse.ArgumentParser(description="Lightbus management command.")

    subparsers = parser.add_subparsers(help="Commands", dest="subcommand")
    subparsers.required = True

    # Allow each command to set up its own arguments
    lightbus.commands.run.Command().setup(parser, subparsers)
    lightbus.commands.shell.Command().setup(parser, subparsers)
    lightbus.commands.dump_schema.Command().setup(parser, subparsers)
    lightbus.commands.dump_config_schema.Command().setup(parser, subparsers)
    lightbus.commands.inspect.Command().setup(parser, subparsers)
    lightbus.commands.version.Command().setup(parser, subparsers)

    # Create a temporary plugin registry in order to run the before_parse_args hook
    plugin_registry = PluginRegistry()
    plugin_registry.autoload_plugins(config=Config.load_dict({}))

    block(
        plugin_registry.execute_hook("before_parse_args", parser=parser, subparsers=subparsers),
        timeout=5,
    )
    args = parser.parse_args(sys.argv[1:] if args is None else args)
    # Note that we don't have an after_parse_args plugin hook. Instead we use the receive_args
    # hook which is called once we have instantiated our plugins

    return args


def load_config(args) -> Config:
    return lightbus.creation.load_config(
        from_file=args.config_file,
        service_name=args.service_name,
        process_name=args.process_name,
        quiet=True,
    )



import json
import logging
from typing import Mapping, Dict, Optional

from lightbus.transports.base import SchemaTransport
from lightbus.schema.encoder import json_encode
from lightbus.transports.redis.utilities import RedisTransportMixin
from lightbus.utilities.frozendict import frozendict

logger = logging.getLogger("lightbus.transports.redis")


class RedisSchemaTransport(RedisTransportMixin, SchemaTransport):
    def __init__(
        self,
        *,
        redis_pool=None,
        url: str = "redis://127.0.0.1:6379/0",
        connection_parameters: Mapping = frozendict(),
    ):
        self.set_redis_pool(redis_pool, url, connection_parameters)
        self._latest_ids = {}
        super().__init__()

    @classmethod
    def from_config(
        cls,
        config,
        url: str = "redis://127.0.0.1:6379/0",
        connection_parameters: Mapping = frozendict(),
    ):
        return cls(url=url, connection_parameters=connection_parameters)

    def schema_key(self, api_name):
        return "schema:{}".format(api_name)

    def schema_set_key(self):
        """Maintains a set of api names in redis which can be used to retrieve individual schemas"""
        return "schemas"

    async def store(self, api_name: str, schema: Dict, ttl_seconds: Optional[int]):
        """Store an individual schema"""
        with await self.connection_manager() as redis:
            schema_key = self.schema_key(api_name)

            p = redis.pipeline()
            p.set(schema_key, json_encode(schema))
            if ttl_seconds is not None:
                p.expire(schema_key, ttl_seconds)
            p.sadd(self.schema_set_key(), api_name)
            await p.execute()

    async def load(self) -> Dict[str, Dict]:
        """Load all schemas"""
        schemas = {}
        with await self.connection_manager() as redis:
            # Get & decode the api names
            api_names = list(await redis.smembers(self.schema_set_key()))
            api_names = [api_name.decode("utf8") for api_name in api_names]

            # Convert the api names into redis keys
            keys = [self.schema_key(api_name) for api_name in api_names]

            if not keys:
                return {}

            # Get the schemas from the keys
            encoded_schemas = await redis.mget(*keys)
            for api_name, schema in zip(api_names, encoded_schemas):
                # Schema may have expired
                if schema:
                    schemas[api_name] = json.loads(schema)
        return schemas



from itertools import chain
from typing import NamedTuple, Dict, Sequence, List, Set, Type, Union, TYPE_CHECKING, Tuple

from lightbus.exceptions import TransportNotFound, TransportsNotInstalled
from lightbus.transports.pool import TransportPool
from lightbus.utilities.importing import load_entrypoint_classes

empty = NamedTuple("Empty")

if TYPE_CHECKING:
    # pylint: disable=unused-import,cyclic-import
    from lightbus.config import Config
    from lightbus.transports import (
        RpcTransport,
        ResultTransport,
        EventTransport,
        SchemaTransport,
        Transport,
    )

EventTransportPoolType = TransportPool["EventTransport"]
RpcTransportPoolType = TransportPool["RpcTransport"]
ResultTransportPoolType = TransportPool["ResultTransport"]
SchemaTransportPoolType = TransportPool["SchemaTransport"]
AnyTransportPoolType = Union[
    EventTransportPoolType, RpcTransportPoolType, ResultTransportPoolType, SchemaTransportPoolType
]


class TransportRegistry:
    """ Manages access to transports

    It is possible for different APIs within lightbus to use different transports.
    This registry handles the logic of loading the transports for a given
    configuration. Thereafter, it provides access to these transports based on
    a given API.

    The 'default' API is a special case as it is fallback transport for
    any APIs that do not have their own specific transports configured.
    """

    class _RegistryEntry(NamedTuple):
        rpc: RpcTransportPoolType = None
        result: ResultTransportPoolType = None
        event: EventTransportPoolType = None

    schema_transport: TransportPool = None

    def __init__(self):
        self._registry: Dict[str, TransportRegistry._RegistryEntry] = {}

    def load_config(self, config: "Config") -> "TransportRegistry":
        # For every configured API...
        for api_name, api_config in config.apis().items():
            # ...and for each type of transport...
            for transport_type in ("event", "rpc", "result"):
                # ...get the transport config...
                transport_selector = getattr(api_config, f"{transport_type}_transport")
                transport_config = self._get_transport_config(transport_selector)
                # ... and use it to create the transport.
                if transport_config:
                    transport_name, transport_config = transport_config
                    transport_class = get_transport(type_=transport_type, name=transport_name)
                    self._set_transport(
                        api_name, transport_class, transport_type, transport_config, config
                    )

        # Schema transport
        transport_config = self._get_transport_config(config.bus().schema.transport)
        if transport_config:
            transport_name, transport_config = transport_config
            transport_class = get_transport(type_="schema", name=transport_name)
            self.schema_transport = self._instantiate_transport_pool(
                transport_class, transport_config, config
            )

        return self

    def _get_transport_config(self, transport_selector):
        if transport_selector:
            for transport_name in transport_selector._fields:
                transport_config = getattr(transport_selector, transport_name)
                if transport_config is not None:
                    return transport_name, transport_config

    def _instantiate_transport_pool(
        self, transport_class: Type["Transport"], transport_config: NamedTuple, config: "Config"
    ):
        transport_pool = TransportPool(
            transport_class=transport_class, transport_config=transport_config, config=config
        )
        return transport_pool

    def _set_transport(
        self,
        api_name: str,
        transport_class: Type["Transport"],
        transport_type: str,
        transport_config: NamedTuple,
        config: "Config",
    ):
        """Set the transport pool for a specific API"""
        from lightbus.transports import Transport

        assert issubclass(
            transport_class, Transport
        ), f"Must be a subclass for Transport, was {transport_class}"

        self._registry.setdefault(api_name, self._RegistryEntry())
        transport_pool = self._instantiate_transport_pool(transport_class, transport_config, config)
        self._registry[api_name] = self._registry[api_name]._replace(
            **{transport_type: transport_pool}
        )

    def _get_transport_pool(
        self, api_name: str, transport_type: str, default=empty
    ) -> AnyTransportPoolType:
        # Get the registry entry for this API (if any)
        registry_entry = self._registry.get(api_name)
        api_transport = None

        # If we have a registry entry for this API, then get the transport for it
        if registry_entry:
            api_transport = getattr(registry_entry, transport_type)

        # Otherwise get the transport for the default API (which is always our fallback)
        # (but don't bother if they have explicity asked for the default_api, as if they
        # have then we've already failed to get that in the previous step)
        if not api_transport and api_name != "default":
            try:
                api_transport = self._get_transport_pool("default", transport_type)
            except TransportNotFound:
                pass

        # If we STILL don't have a transport then show a sensible error
        if not api_transport and default == empty:
            raise TransportNotFound(
                f"No {transport_type} transport found for API '{api_name}'. Neither was a default "
                f"API transport found. Either specify a {transport_type} transport for this specific API, "
                f"or specify a default {transport_type} transport. In most cases setting a default transport "
                f"is the best course of action."
            )
        else:
            return api_transport

    def _get_transport_pools(
        self, api_names: Sequence[str], transport_type: str
    ) -> Dict[AnyTransportPoolType, List[str]]:
        apis_by_transport: Dict[AnyTransportPoolType, List[str]] = {}
        for api_name in api_names:
            transport = self._get_transport_pool(api_name, transport_type)
            apis_by_transport.setdefault(transport, [])
            apis_by_transport[transport].append(api_name)
        return apis_by_transport

    def _has_transport(self, api_name: str, transport_type: str) -> bool:
        try:
            self._get_transport_pool(api_name, transport_type)
        except TransportNotFound:
            return False
        else:
            return True

    def set_rpc_transport(
        self,
        api_name: str,
        transport_class: Type["RpcTransport"],
        transport_config: NamedTuple,
        config: "Config",
    ):
        self._set_transport(api_name, transport_class, "rpc", transport_config, config)

    def set_result_transport(
        self,
        api_name: str,
        transport_class: Type["ResultTransport"],
        transport_config: NamedTuple,
        config: "Config",
    ):
        self._set_transport(api_name, transport_class, "result", transport_config, config)

    def set_event_transport(
        self,
        api_name: str,
        transport_class: Type["EventTransport"],
        transport_config: NamedTuple,
        config: "Config",
    ):
        self._set_transport(api_name, transport_class, "event", transport_config, config)

    def set_schema_transport(
        self,
        transport_class: Type["SchemaTransport"],
        transport_config: NamedTuple,
        config: "Config",
    ):
        self.schema_transport = self._instantiate_transport_pool(
            transport_class, transport_config, config
        )

    def get_rpc_transport(self, api_name: str, default=empty) -> RpcTransportPoolType:
        return self._get_transport_pool(api_name, "rpc", default=default)

    def get_result_transport(self, api_name: str, default=empty) -> ResultTransportPoolType:
        return self._get_transport_pool(api_name, "result", default=default)

    def get_event_transport(self, api_name: str, default=empty) -> EventTransportPoolType:
        return self._get_transport_pool(api_name, "event", default=default)

    def get_all_rpc_transports(self) -> Set[RpcTransportPoolType]:
        return {t.rpc for t in self._registry.values() if t.rpc}

    def get_all_result_transports(self) -> Set[ResultTransportPoolType]:
        return {t.result for t in self._registry.values() if t.result}

    def get_all_event_transports(self) -> Set[EventTransportPoolType]:
        return {t.event for t in self._registry.values() if t.event}

    def get_schema_transport(self, default=empty) -> SchemaTransportPoolType:
        if self.schema_transport or default != empty:
            return self.schema_transport or default
        else:
            # TODO: Link to docs
            raise TransportNotFound(
                "No schema transport is configured for this bus. Check your schema transport "
                "configuration is setup correctly (config section: bus.schema.transport)."
            )

    def has_rpc_transport(self, api_name: str) -> bool:
        return self._has_transport(api_name, "rpc")

    def has_result_transport(self, api_name: str) -> bool:
        return self._has_transport(api_name, "result")

    def has_event_transport(self, api_name: str) -> bool:
        return self._has_transport(api_name, "event")

    def has_schema_transport(self) -> bool:
        return bool(self.schema_transport)

    def get_rpc_transports(self, api_names: Sequence[str]) -> Dict[RpcTransportPoolType, List[str]]:
        """Get a mapping of transports to lists of APIs

        This is useful when multiple APIs can be served by a single transport
        """
        return self._get_transport_pools(api_names, "rpc")

    def get_event_transports(
        self, api_names: Sequence[str]
    ) -> Dict[EventTransportPoolType, List[str]]:
        """Get a mapping of transports to lists of APIs

        This is useful when multiple APIs can be served by a single transport
        """
        return self._get_transport_pools(api_names, "event")

    def get_all_transports(self) -> Set[AnyTransportPoolType]:
        """Get a set of all transports irrespective of type"""
        all_transports = chain(*[entry._asdict().values() for entry in self._registry.values()])
        return set([t for t in all_transports if t is not None])


def get_available_transports(type_):
    loaded = load_entrypoint_classes(f"lightbus_{type_}_transports")

    if not loaded:
        raise TransportsNotInstalled(
            f"No {type_} transports are available, which means lightbus has not been "
            f"installed correctly. This is likely because you are working on Lightbus itself. "
            f"In which case, within your local lightbus repo you should run "
            f"something like 'pip install .' or 'python setup.py develop'.\n\n"
            f"This will install the entrypoints (defined in setup.py) which point Lightbus "
            f"to it's bundled transports."
        )
    return {name: class_ for module_name, name, class_ in loaded}


def get_transport(type_, name):
    for name_, class_ in get_available_transports(type_).items():
        if name == name_:
            return class_

    raise TransportNotFound(
        f"No '{type_}' transport found named '{name}'. Check the transport is installed and "
        f"has the relevant entrypoints setup in it's setup.py file. Or perhaps "
        f"you have a typo in your config file."
    )


def get_transport_name(cls: Type[AnyTransportPoolType]):
    for type_ in ("rpc", "result", "event"):
        for *_, name, class_ in load_entrypoint_classes(f"lightbus_{type_}_transports"):
            if cls == class_:
                return name

    raise TransportNotFound(
        f"Transport class {cls.__module__}.{cls.__name__} is not specified in any entrypoint."
    )



import asyncio
import logging

from lightbus.client.docks.base import BaseDock
from lightbus.client.utilities import queue_exception_checker
from lightbus.client import commands
from lightbus.utilities.async_tools import cancel
from lightbus.utilities.singledispatch import singledispatchmethod

logger = logging.getLogger(__name__)


class EventDock(BaseDock):
    """ Takes internal Lightbus commands and performs interactions with the Event transport
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.listener_tasks = set()

    @singledispatchmethod
    async def handle(self, command):
        raise NotImplementedError(f"Did not recognise command {command.__class__.__name__}")

    @handle.register
    async def handle_consume_events(self, command: commands.ConsumeEventsCommand):
        event_transports = self.transport_registry.get_event_transports(
            api_names=[api_name for api_name, _ in command.events]
        )

        async def listener(event_transport, events_):
            consumer = event_transport.consume(
                listen_for=events_,
                listener_name=command.listener_name,
                error_queue=self.error_queue,
                **command.options,
            )
            async for event_messages in consumer:
                for event_message in event_messages:
                    logger.debug(f"Putting event {event_message.id} onto the internal queue")
                    await command.destination_queue.put(event_message)

                # Wait for the queue to be emptied before fetching more.
                # We will need to make this configurable if we want to support
                # the pre-fetching of events. This solution is a good default
                # though as it will ensure events are processed in a more ordered fashion
                # in cases where there are multiple workers, and also ensure fewer
                # messages need to be reclaimed in the event of a crash
                logger.debug(f"Waiting for messages to be consumed before adding more")
                await command.destination_queue.join()

        for event_transport_pool, api_names in event_transports.items():
            # Create a listener task for each event transport,
            # passing each a list of events for which it should listen
            events = [
                (api_name, event_name)
                for api_name, event_name in command.events
                if api_name in api_names
            ]

            # fmt: off
            listener_task = asyncio.ensure_future(queue_exception_checker(
                listener(event_transport_pool, events),
                self.error_queue,
            ))
            # fmt: on
            self.listener_tasks.add(listener_task)

    @handle.register
    async def handle_send_event(self, command: commands.SendEventCommand):
        event_transport = self.transport_registry.get_event_transport(command.message.api_name)

        await event_transport.send_event(event_message=command.message, options=command.options)

    @handle.register
    async def handle_acknowledge_event(self, command: commands.AcknowledgeEventCommand):
        event_transport = self.transport_registry.get_event_transport(command.message.api_name)
        await event_transport.acknowledge(command.message)

    @handle.register
    async def handle_close(self, command: commands.CloseCommand):
        await cancel(*self.listener_tasks)

        for event_transport in self.transport_registry.get_all_event_transports():
            await event_transport.close()

        await self.consumer.close()
        await self.producer.close()



import inspect
import logging
import secrets
import string

from typing import Type, NamedTuple  # pylint: disable=unused-import,cyclic-import

import itertools

logger = logging.getLogger(__name__)


def make_from_config_structure(class_name, from_config_method, extra_parameters=tuple()) -> Type:
    """
    Create a new named tuple based on the method signature of from_config_method.

    This is useful when dynamically creating the config structure for Transports
    and Plugins.
    """
    # pylint: disable=exec-used

    code = f"class {class_name}Config(NamedTuple):\n    pass\n"
    variables = dict(p={})

    parameters = inspect.signature(from_config_method).parameters.values()
    for parameter in itertools.chain(parameters, extra_parameters):
        if parameter.name == "config":
            # The config parameter is always passed to from_config() in order to
            # give it access to the global configuration (useful for setting
            # sensible defaults)
            continue

        if parameter.kind in (parameter.POSITIONAL_ONLY, parameter.VAR_POSITIONAL):
            logger.warning(
                f"Positional-only arguments are not supported in from_config() on class {class_name}"
            )
        elif parameter.kind in (parameter.VAR_KEYWORD,):
            logger.warning(
                f"**kwargs-style parameters are not supported in from_config() on class {class_name}"
            )
        else:
            name = parameter.name
            variables["p"][name] = parameter
            code += f"    {name}: p['{name}'].annotation = p['{name}'].default\n"

    globals_ = globals().copy()
    globals_.update(variables)
    exec(code, globals_)  # nosec
    return globals_[f"{class_name}Config"]


def random_name(length: int) -> str:
    """Get a random string suitable for a processes/consumer name"""
    return "".join(secrets.choice(string.ascii_lowercase) for _ in range(length))



import importlib
import logging
import sys
from typing import Sequence, Tuple, Callable

import pkg_resources

logger = logging.getLogger(__name__)


def import_module_from_string(name):
    """Import a module if necessary, otherwise return it from the list of already imported modules"""
    if name in sys.modules:
        return sys.modules[name]
    else:
        return importlib.import_module(name)


def import_from_string(name):
    components = name.split(".")
    mod = __import__(components[0])
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


def load_entrypoint_classes(entrypoint_name) -> Sequence[Tuple[str, str, Callable]]:
    """Load classes specified in an entrypoint

    Entrypoints are specified in setup.py, and Lightbus uses them to
    discover plugins & transports.
    """
    found_classes = []
    for entrypoint in pkg_resources.iter_entry_points(entrypoint_name):
        class_ = entrypoint.load()
        found_classes.append((entrypoint.module_name, entrypoint.name, class_))
    return found_classes



import logging
import time
from typing import Mapping, TYPE_CHECKING

from lightbus.transports.base import ResultTransport, ResultMessage, RpcMessage
from lightbus.log import L, Bold
from lightbus.serializers import BlobMessageSerializer, BlobMessageDeserializer
from lightbus.transports.redis.utilities import RedisTransportMixin
from lightbus.utilities.frozendict import frozendict
from lightbus.utilities.human import human_time
from lightbus.utilities.importing import import_from_string

if TYPE_CHECKING:
    # pylint: disable=unused-import,cyclic-import
    from lightbus.config import Config
    from lightbus.client import BusClient

logger = logging.getLogger("lightbus.transports.redis")


class RedisResultTransport(RedisTransportMixin, ResultTransport):
    """ Redis result transport

    For a description of the protocol see https://lightbus.org/reference/protocols/rpc-and-result/
    """

    def __init__(
        self,
        *,
        redis_pool=None,
        url=None,
        serializer=BlobMessageSerializer(),
        deserializer=BlobMessageDeserializer(ResultMessage),
        connection_parameters: Mapping = frozendict(maxsize=100),
        result_ttl=60,
        rpc_timeout=5,
    ):
        # NOTE: We use the blob message_serializer here, as the results come back as single values in a redis list
        self.set_redis_pool(redis_pool, url, connection_parameters)
        self.serializer = serializer
        self.deserializer = deserializer
        self.result_ttl = result_ttl
        self.rpc_timeout = rpc_timeout
        super().__init__()

    @classmethod
    def from_config(
        cls,
        config: "Config",
        url: str = "redis://127.0.0.1:6379/0",
        serializer: str = "lightbus.serializers.BlobMessageSerializer",
        deserializer: str = "lightbus.serializers.BlobMessageDeserializer",
        connection_parameters: Mapping = frozendict(maxsize=100),
        result_ttl=60,
        rpc_timeout=5,
    ):
        serializer = import_from_string(serializer)()
        deserializer = import_from_string(deserializer)(ResultMessage)

        return cls(
            url=url,
            serializer=serializer,
            deserializer=deserializer,
            connection_parameters=connection_parameters,
            result_ttl=result_ttl,
            rpc_timeout=rpc_timeout,
        )

    async def get_return_path(self, rpc_message: RpcMessage) -> str:
        """Get the return path for the given message

        The return path is sent with the outgoing RPC message and
        tells the received where the caller expects to find the result.

        In this case, the return patch specifies a specific key in redis
        in which the results should be placed.
        """
        return "redis+key://{}.{}:result:{}".format(
            rpc_message.api_name, rpc_message.procedure_name, rpc_message.id
        )

    def _parse_return_path(self, return_path: str) -> str:
        """Get the redis key specified by the given return path"""
        if not return_path.startswith("redis+key://"):
            raise AssertionError(f"Invalid return path specified: {return_path}")
        return return_path[12:]

    async def send_result(
        self, rpc_message: RpcMessage, result_message: ResultMessage, return_path: str
    ):
        """Send the result back to the caller"""
        logger.debug(
            L(
                "Sending result {} into Redis using return path {}",
                Bold(result_message),
                Bold(return_path),
            )
        )
        redis_key = self._parse_return_path(return_path)

        with await self.connection_manager() as redis:
            start_time = time.time()
            p = redis.pipeline()
            p.lpush(redis_key, self.serializer(result_message))
            p.expire(redis_key, timeout=self.result_ttl)
            await p.execute()

        logger.debug(
            L(
                "➡ Sent result {} into Redis in {} using return path {}",
                Bold(result_message),
                human_time(time.time() - start_time),
                Bold(return_path),
            )
        )

    async def receive_result(
        self, rpc_message: RpcMessage, return_path: str, options: dict
    ) -> ResultMessage:
        """Await a result from the processing worker"""
        logger.debug(L("Awaiting Redis result for RPC message: {}", Bold(rpc_message)))
        redis_key = self._parse_return_path(return_path)

        with await self.connection_manager() as redis:
            start_time = time.time()
            result = None
            while not result:
                # Sometimes blpop() will return None in the case of timeout or
                # cancellation. We therefore perform this step with a loop to catch
                # this. A more elegant solution is welcome.
                # TODO: RPC & result Transports should not be applying a timeout, leave
                #       this to the client which coordinates between the two
                result = await redis.blpop(redis_key, timeout=self.rpc_timeout)
            _, serialized = result

        result_message = self.deserializer(
            serialized, api_name=rpc_message.api_name, procedure_name=rpc_message.procedure_name
        )

        logger.debug(
            L(
                "⬅ Received Redis result in {} for RPC message {}: {}",
                human_time(time.time() - start_time),
                rpc_message,
                Bold(result_message.result),
            )
        )

        return result_message



# The opposite of deforming. See lightbus.utilities.deforming
import datetime
import inspect
import logging
from enum import Enum
from typing import Mapping, Type, get_type_hints, Union, TypeVar, Callable, Any
from base64 import b64decode

from lightbus.utilities.type_checks import (
    type_is_namedtuple,
    type_is_dataclass,
    is_optional,
    isinstance_safe,
    parse_hint,
    issubclass_safe,
)

is_callable = callable
logger = logging.getLogger(__name__)


def cast_to_signature(parameters: dict, callable) -> dict:
    """Cast parameters into the type hints provided by callable's function signature.

    Will resolve functions which have been wrapped using itertools.wrap()
    """
    casted_parameters = parameters.copy()
    for key, hint in get_type_hints(callable).items():
        if key not in casted_parameters:
            continue

        casted_parameters[key] = cast_to_hint(value=casted_parameters[key], hint=hint)

    return casted_parameters


V = TypeVar("V")
H = TypeVar("A")


def cast_to_hint(value: V, hint: H) -> Union[V, H]:
    """Cast a value into a given type hint

    If a value cannot be cast then the original value will be returned
    and a warning emitted
    """
    # pylint: disable=too-many-return-statements
    if value is None:
        return None
    elif hint in (Any, ...):
        return value

    optional_hint = is_optional(hint)
    if optional_hint and value is not None:
        hint = optional_hint

    hint_type, hint_args = parse_hint(hint)
    is_class = inspect.isclass(hint_type)

    if hint_type == Union:
        # We don't attempt to deal with unions for now
        return value
    elif hint == inspect.Parameter.empty:
        # Empty annotation
        return value
    elif isinstance_safe(value, hint):
        # Already correct type
        return value
    elif hasattr(hint, "__from_bus__"):
        # Hint supports custom deserializing.
        return _mapping_to_instance(
            mapping=value,
            destination_type=hint,
            instantiator=hint.__from_bus__,
            expand_kwargs=False,
        )
    elif issubclass_safe(hint, bytes):
        return b64decode(value.encode("utf8"))
    elif type_is_namedtuple(hint) and isinstance_safe(value, Mapping):
        return _mapping_to_instance(mapping=value, destination_type=hint)
    elif type_is_dataclass(hint) and isinstance_safe(value, Mapping):
        return _mapping_to_instance(mapping=value, destination_type=hint)
    elif is_class and issubclass_safe(hint_type, datetime.datetime) and isinstance_safe(value, str):
        # Datetime as a string
        return datetime.datetime.fromisoformat(value)
    elif is_class and issubclass_safe(hint_type, datetime.date) and isinstance_safe(value, str):
        # Date as a string
        return cast_or_warning(lambda v: datetime.datetime.fromisoformat(v).date(), value)
    elif is_class and issubclass_safe(hint_type, list):
        # Lists
        if hint_args and hasattr(value, "__iter__"):
            value = [cast_to_hint(i, hint_args[0]) for i in value]
        return cast_or_warning(list, value)
    elif is_class and issubclass_safe(hint_type, tuple):
        # Tuples
        if hint_args and hasattr(value, "__iter__"):
            value = [cast_to_hint(h, hint_args[i]) for i, h in enumerate(value)]
        return cast_or_warning(tuple, value)
    elif is_class and issubclass_safe(hint_type, set):
        # Sets
        if hint_args and hasattr(value, "__iter__"):
            value = [cast_to_hint(i, hint_args[0]) for i in value]
        return cast_or_warning(set, value)
    elif (
        inspect.isclass(hint)
        and hasattr(hint, "__annotations__")
        and not issubclass_safe(hint_type, Enum)
    ):
        logger.warning(
            f"Cannot cast to arbitrary class {hint}, using un-casted value. "
            f"If you want to receive custom objects you can 1) "
            f"use a NamedTuple, 2) use a dataclass, or 3) specify the "
            f"__from_bus__() and __to_bus__() magic methods."
        )
        return value
    else:
        return cast_or_warning(hint, value)


T = TypeVar("T")


def _mapping_to_instance(
    mapping: Mapping, destination_type: Type[T], instantiator: Callable = None, expand_kwargs=True
) -> T:
    """Convert a dictionary-like object into an instance of the given type

    This conversion is performed recursively. If the passed type
    class contains child structures, then the corresponding
    child keys of the dictionary will be mapped.

    If `instantiator` is provided, it will be called rather than simply
    calling `destination_type`.

    If `expand_kwargs` is True then the the parameters will be
    expended (i.e. **) when the `destination_type` / `instantiator`
    is called.
    """
    import lightbus.config.structure

    hints = get_type_hints(destination_type, None, lightbus.config.structure.__dict__)
    parameters = {}

    if mapping is None:
        return None

    # Iterate through each key/type-hint pairing in the destination type
    for key, hint in hints.items():
        value = mapping.get(key)

        if key not in mapping:
            # This attribute has not been provided by in the mapping. Skip it
            continue

        hint_type, hint_args = parse_hint(hint)

        # Is this an Optional[] hint (which looks like Union[Thing, None])
        if is_optional(hint) and value is not None:
            # Yep, it's an Optional. So unwrap it.
            hint = hint_args[0]

        if issubclass_safe(hint_type, Mapping) and hint_args and len(hint_args) == 2:
            parameters[key] = dict()
            for k, v in value.items():
                parameters[key][k] = cast_to_hint(v, hint_args[1])
        else:
            parameters[key] = cast_to_hint(value, hint)

    instantiator = instantiator or destination_type
    if expand_kwargs:
        return instantiator(**parameters)
    else:
        return instantiator(parameters)


def cast_or_warning(type_, value):
    try:
        return type_(value)
    except Exception as e:
        logger.warning(
            f"Failed to cast value {repr(value)} to type {type_}. Will "
            f"continue without casting, but this may cause errors in any "
            f"called code. Error was: {e}"
        )
        return value



import logging
from logging import Logger

from typing import TYPE_CHECKING

from lightbus.config import Config
from lightbus.log import LightbusFormatter, LBullets, L, Bold
from lightbus.plugins import PluginRegistry
from lightbus.transports.registry import TransportRegistry

if TYPE_CHECKING:
    # pylint: disable=unused-import,cyclic-import
    from lightbus import RpcTransport, ResultTransport, EventTransport, SchemaTransport, Schema

handler = logging.StreamHandler()


def configure_logging(log_level=logging.INFO):
    logger = logging.getLogger("lightbus")
    logger.propagate = False

    formatter = LightbusFormatter()
    handler.setFormatter(formatter)

    logger.removeHandler(handler)
    logger.addHandler(handler)
    logger.setLevel(log_level)


def log_welcome_message(
    logger: Logger,
    transport_registry: TransportRegistry,
    schema: "Schema",
    plugin_registry: PluginRegistry,
    config: Config,
):
    """Show the worker-startup welcome message
    """
    logger.info(
        LBullets(
            "Lightbus is setting up",
            items={
                "service_name (set with -s or LIGHTBUS_SERVICE_NAME)": Bold(config.service_name),
                "process_name (with with -p or LIGHTBUS_PROCESS_NAME)": Bold(config.process_name),
            },
        )
    )

    # Log the transport information
    rpc_transport = transport_registry.get_rpc_transport("default", default=None)
    result_transport = transport_registry.get_result_transport("default", default=None)
    event_transport = transport_registry.get_event_transport("default", default=None)
    log_transport_information(
        rpc_transport, result_transport, event_transport, schema.schema_transport, logger
    )

    if plugin_registry._plugins:
        logger.info(
            LBullets(
                "Loaded the following plugins ({})".format(len(plugin_registry._plugins)),
                items=plugin_registry._plugins,
            )
        )
    else:
        logger.info("No plugins loaded")


def log_transport_information(
    rpc_transport: "RpcTransport",
    result_transport: "ResultTransport",
    event_transport: "EventTransport",
    schema_transport: "SchemaTransport",
    logger: Logger,
):
    def stringify_transport(transport):
        if not transport:
            return "None"

        # Rely on the transport to display itself sensibly as a string
        return L("{} @ {}", transport.__class__.__name__, Bold(transport))

    logger.info(
        LBullets(
            "Default transports are setup as follows:",
            items={
                "RPC transport": stringify_transport(rpc_transport),
                "Result transport": stringify_transport(result_transport),
                "Event transport": stringify_transport(event_transport),
                "Schema transport": stringify_transport(schema_transport),
            },
        )
    )



import asyncio
import logging
from datetime import datetime
from typing import Sequence, Tuple, List, Dict, AsyncGenerator, TYPE_CHECKING

from lightbus.transports.base import ResultTransport, RpcTransport, EventTransport, SchemaTransport
from lightbus.message import RpcMessage, EventMessage, ResultMessage

if TYPE_CHECKING:
    # pylint: disable=unused-import,cyclic-import
    from lightbus.client import BusClient

logger = logging.getLogger(__name__)


class DebugRpcTransport(RpcTransport):
    async def call_rpc(self, rpc_message: RpcMessage, options: dict):
        """Publish a call to a remote procedure"""
        logger.debug("Faking dispatch of message {}".format(rpc_message))

    async def consume_rpcs(self, apis) -> Sequence[RpcMessage]:
        """Consume RPC calls for the given API"""
        logger.debug("Faking consumption of RPCs. Waiting 100ms before issuing fake RPC call...")
        await asyncio.sleep(0.1)
        logger.debug("Issuing fake RPC call")
        return self._get_fake_messages()

    def _get_fake_messages(self):
        return [
            RpcMessage(
                api_name="my_company.auth",
                procedure_name="check_password",
                kwargs=dict(username="admin", password="secret"),  # nosec
            )
        ]


class DebugResultTransport(ResultTransport):
    async def get_return_path(self, rpc_message: RpcMessage) -> str:
        return "debug://foo"

    async def send_result(
        self, rpc_message: RpcMessage, result_message: ResultMessage, return_path: str
    ):
        logger.info("Faking sending of result: {}".format(result_message))

    async def receive_result(
        self, rpc_message: RpcMessage, return_path: str, options: dict
    ) -> ResultMessage:
        logger.info("⌛ Faking listening for results. Will issue fake result in 0.5 seconds...")
        await asyncio.sleep(0.1)  # This is relied upon in testing
        logger.debug("Faking received result")

        return ResultMessage(
            result="Fake result",
            rpc_message_id=rpc_message.id,
            api_name=rpc_message.api_name,
            procedure_name=rpc_message.procedure_name,
        )


class DebugEventTransport(EventTransport):
    def __init__(self):
        self._task = None
        self._reload = False
        self._events = set()
        super().__init__()

    async def send_event(self, event_message: EventMessage, options: dict):
        """Publish an event"""
        logger.info(
            " Faking sending of event {}.{} with kwargs: {}".format(
                event_message.api_name, event_message.event_name, event_message.kwargs
            )
        )

    async def consume(
        self, listen_for: List[Tuple[str, str]], listener_name: str, **kwargs
    ) -> AsyncGenerator[EventMessage, None]:
        """Consume RPC events for the given API"""
        self._sanity_check_listen_for(listen_for)

        logger.info("⌛ Faking listening for events {}.".format(self._events))

        while True:
            await asyncio.sleep(0.1)
            yield [self._get_fake_message()]

    async def history(
        self,
        api_name,
        event_name,
        start: datetime = None,
        stop: datetime = None,
        start_inclusive: bool = True,
    ) -> AsyncGenerator[EventMessage, None]:
        yield self._get_fake_message()

    def _get_fake_message(self):
        message = EventMessage(
            api_name="my_company.auth", event_name="user_registered", kwargs={"example": "value"}
        )
        setattr(message, "datetime", datetime.now())
        return message


class DebugSchemaTransport(SchemaTransport):
    def __init__(self):
        self._schemas = {}

    async def store(self, api_name: str, schema: Dict, ttl_seconds: int):
        self._schemas[api_name] = schema

        logging.debug(
            "Debug schema transport storing schema for {} (TTL: {}): {}".format(
                api_name, ttl_seconds, schema
            )
        )

    async def load(self) -> Dict[str, Dict]:
        return self._schemas



# pylint: skip-file
"""Lightbus logging setup

Credit to: https://github.com/borntyping/python-colorlog

This has been vendored to minimise dependencies and allow for
customisation in the long term.

This needs some serious cleanup: https://github.com/adamcharnock/lightbus/issues/10
"""

import logging
import sys

__all__ = ("escape_codes", "default_log_colors", "LightbusFormatter")

# The default colors to use for the debug levels
default_log_colors = {
    "DEBUG": "white",
    "INFO": "green",
    "WARNING": "yellow",
    "ERROR": "red",
    "CRITICAL": "bold_red",
}

# The default format to use for each style
default_formats = {
    "%": {
        "DEBUG": "%(log_color)s%(threadName)s | %(name)s | %(msg)s",
        "INFO": "%(log_color)s%(threadName)s | %(name)s | %(msg)s",
        "WARNING": "%(log_color)s%(threadName)s | %(name)s | %(msg)s",
        "ERROR": "%(log_color)s%(threadName)s | %(name)s | ERROR: %(msg)s (%(module)s:%(lineno)d)",
        "CRITICAL": "%(log_color)s%(threadName)s | %(name)s | CRITICAL: %(msg)s",
    },
    "{": "{log_color} {levelname}:{name}:{message}",
    "$": "${log_color}$ ${levelname}:${name}:${message}",
}


def esc(*x):
    return "\033[" + ";".join(x) + "m"


escape_codes = {"reset": esc("0"), "bold": esc("01"), "thin": esc("02")}

# The color names
COLORS = ["black", "red", "green", "yellow", "blue", "purple", "cyan", "white"]

PREFIXES = [
    # Foreground without prefix
    ("3", ""),
    ("01;3", "bold_"),
    ("02;3", "thin_"),
    # Foreground with fg_ prefix
    ("3", "fg_"),
    ("01;3", "fg_bold_"),
    ("02;3", "fg_thin_"),
    # Background with bg_ prefix - bold/light works differently
    ("4", "bg_"),
    ("10", "bg_bold_"),
]

for prefix, prefix_name in PREFIXES:
    for code, name in enumerate(COLORS):
        escape_codes[prefix_name + name] = esc(prefix + str(code))


def parse_colors(sequence):
    """Return escape codes from a color sequence."""
    return "".join(escape_codes[n] for n in sequence.split(",") if n)


class LightbusLogRecord:  # pragma: no cover
    """
    Wraps a LogRecord, adding named escape codes to the internal dict.

    The internal dict is used when formatting the message (by the PercentStyle,
    StrFormatStyle, and StringTemplateStyle classes).
    """

    def __init__(self, record):
        """Add attributes from the escape_codes dict and the record."""
        self.is_tty = True
        self.additional_line_prefix = ""
        self.msg_cb = None
        record.name = record.name.ljust(30)

        self.__dict__.update(escape_codes)
        self.__dict__.update(record.__dict__)

        # Keep a reference to the original record so ``__getattr__`` can
        # access functions that are not in ``__dict__``
        self._record = record

    def getMessage(self):
        # Render the message if necessary (i.e. because it is an instance of the L class)
        if hasattr(self.msg, "render"):
            self.msg = self.msg.render(
                tty=self.is_tty,
                additional_line_prefix=self.additional_line_prefix,
                style=self.log_color,
                msg_cb=self.msg_cb,
            )
        else:
            if self.msg_cb:
                self.msg = self.msg_cb(self.msg, is_tty=self.is_tty)

        return self._record.getMessage()

    def __getattr__(self, name):
        return getattr(self._record, name)


class LightbusFormatter(logging.Formatter):  # pragma: no cover
    """
    A formatter that allows colors to be placed in the format string.

    Intended to help in creating more readable logging output.
    """

    def __init__(
        self,
        fmt=None,
        datefmt=None,
        style="%",
        log_colors=None,
        reset=True,
        secondary_log_colors=None,
        stream=None,
    ):
        """
        Set the format and colors the LightbusFormatter will use.

        The ``fmt``, ``datefmt`` and ``style`` args are passed on to the
        ``logging.Formatter`` constructor.

        The ``secondary_log_colors`` argument can be used to create additional
        ``log_color`` attributes. Each key in the dictionary will set
        ``{key}_log_color``, using the value to select from a different
        ``log_colors`` set.

        :Parameters:
        - fmt (str): The format string to use
        - datefmt (str): A format string for the date
        - log_colors (dict):
            A mapping of log level names to color names
        - reset (bool):
            Implicitly append a color reset to all records unless False
        - style ('%' or '{' or '$'):
            The format style to use. (*No meaning prior to Python 3.2.*)
        - secondary_log_colors (dict):
            Map secondary ``log_color`` attributes. (*New in version 2.6.*)
        """
        if fmt is None:
            fmt = default_formats[style]

        if sys.version_info >= (3, 8):
            # Disable the new validation feature in 3.8, as out dictionary
            # of formats will cause it to error. All this logging needs refactoring
            # anyway.
            extra = dict(validate=False)
        else:
            extra = dict()

        super(LightbusFormatter, self).__init__(fmt, datefmt, style, **extra)

        # Disable reset codes if we do not have a TTY
        self.stream = stream or sys.stdout
        reset = reset and self.stream.isatty()

        self.log_colors = log_colors if log_colors is not None else default_log_colors
        self.secondary_log_colors = secondary_log_colors
        self.reset = reset
        self.style = style
        self.fmt = fmt

    def color(self, log_colors, level_name):
        """Return escape codes from a ``log_colors`` dict."""
        # Don't color log records if do not have a TTY
        if not self.stream.isatty():
            log_colors = {}
        return parse_colors(log_colors.get(level_name, ""))

    def format(self, record):
        """Format a message from a record object."""
        record = LightbusLogRecord(record)
        record.log_color = self.color(self.log_colors, record.levelname)
        record.is_tty = self.stream.isatty()

        # Set secondary log colors
        if self.secondary_log_colors:
            for name, log_colors in self.secondary_log_colors.items():
                color = self.color(log_colors, record.levelname)
                setattr(record, name + "_log_color", color)

        # Set format for this particular log level
        if isinstance(self.fmt, dict):
            self._fmt = self.fmt[record.levelname]
            # Update self._style because we've changed self._fmt
            # (code based on stdlib's logging.Formatter.__init__())
            if self.style not in logging._STYLES:
                raise ValueError("Style must be one of: %s" % ",".join(logging._STYLES.keys()))
            self._style = logging._STYLES[self.style][0](self._fmt)

        # Format the message
        record.additional_line_prefix = self.get_additional_line_prefix(record)
        message = super(LightbusFormatter, self).format(record)

        # Add a reset code to the end of the message
        # (if it wasn't explicitly added in format str)
        if self.reset and not message.endswith(escape_codes["reset"]):
            message += escape_codes["reset"]

        return message

    def get_additional_line_prefix(self, record):
        fmt_before_msg = self._fmt.split("%(msg)s", 1)
        if len(fmt_before_msg) == 1:
            return 0
        formatted_prefix = fmt_before_msg[0] % {
            k: "" if "color" in k else " " * len(str(v)) for k, v in record.__dict__.items()
        }
        return formatted_prefix


class L:  # pragma: no cover
    style = ""

    def __init__(self, log_message, *values):
        self.log_message = log_message
        self.values = values

    def __str__(self):
        return self.render()

    def __repr__(self):
        return repr(self.__str__())

    def render(self, parent_style="", style="", tty=True, additional_line_prefix="", msg_cb=None):
        style = style or self.style
        keys = [
            v.render(parent_style=style, tty=tty) if hasattr(v, "render") else v
            for v in self.values
        ]
        if tty:
            msg = str(self.log_message)
            if keys:
                msg = str(self.log_message).format(*keys)
            if msg_cb:
                msg = msg_cb(msg, is_tty=tty)
            return style + msg + escape_codes["reset"] + parent_style
        else:
            return str(self.log_message).format(*keys)


class Bold(L):  # pragma: no cover
    style = escape_codes["bold"]


class LBullets(L):  # pragma: no cover
    def __init__(self, log_message, *values, items, bullet="∙", indent=4):
        super().__init__(log_message, *values)
        self.items = items
        self.bullet = bullet
        self.indent = indent

    def render(self, parent_style="", style="", tty=True, additional_line_prefix="", msg_cb=None):
        if not tty:
            if isinstance(self.items, dict):
                return "{}: {}".format(
                    self.log_message, ", ".join(f"{k}:{v}" for k, v in self.items.items())
                )
            else:
                return "{}: {}".format(self.log_message, ", ".join(map(str, self.items)))

        style = style or self.style

        def render_child(item) -> str:
            if hasattr(item, "render"):
                return item.render(parent_style=style, tty=tty)
            else:
                return item

        rendered_items = []
        if isinstance(self.items, dict):
            key_width = max(map(len, self.items.keys())) + 1
            for k, v in self.items.items():
                rendered_items.append(
                    "{}: {}".format(render_child(k).ljust(key_width), render_child(v))
                )
        else:
            for item in self.items:
                rendered_items.append(render_child(item))

        indent = self.indent
        msg = "\n".join(
            ["{}:".format(self.log_message)]
            + [
                "{}{}{} {}".format(additional_line_prefix, " " * indent, self.bullet, item)
                for item in rendered_items
            ]
            + [additional_line_prefix]
        )
        if msg_cb:
            msg = msg_cb(msg, is_tty=tty)
        return style + msg + escape_codes["reset"] + parent_style



import traceback
from typing import Optional, Dict, Any, Sequence
from uuid import uuid1


__all__ = ["Message", "RpcMessage", "ResultMessage", "EventMessage"]


class Message:
    """Base representation of a Lightbus RPC/Result/Event message"""

    required_metadata: Sequence

    def __init__(self, id: str = "", native_id: str = None):
        self.id = id or str(uuid1())
        self.native_id = native_id

    def get_metadata(self) -> dict:
        """Get the non-kwarg fields of this message

        Will be used by the serializers
        """
        raise NotImplementedError()

    def get_kwargs(self) -> dict:
        """Get the kwarg fields of this message

        Will be used by the serializers
        """
        raise NotImplementedError()

    @classmethod
    def from_dict(cls, metadata: dict, kwargs: dict, **extra) -> "Message":
        """Create a message instance given the metadata and kwargs

        Will be used by the serializers
        """
        raise NotImplementedError()


class RpcMessage(Message):
    """Representation of a Lightbus RPC message"""

    required_metadata = ["id", "api_name", "procedure_name", "return_path"]

    def __init__(
        self,
        *,
        api_name: str,
        procedure_name: str,
        kwargs: Optional[dict] = None,
        return_path: Any = None,
        id: str = "",
        native_id: str = None,
    ):
        super().__init__(id, native_id)
        self.api_name = api_name
        self.procedure_name = procedure_name
        self.kwargs = kwargs
        self.return_path = return_path

    def __repr__(self):
        return "<{}: {}>".format(self.__class__.__name__, self)

    def __str__(self):
        return "{}({})".format(
            self.canonical_name, ", ".join("{}={}".format(k, v) for k, v in self.kwargs.items())
        )

    @property
    def canonical_name(self):
        return "{}.{}".format(self.api_name, self.procedure_name)

    def get_metadata(self) -> dict:
        return {
            "id": self.id,
            "api_name": self.api_name,
            "procedure_name": self.procedure_name,
            "return_path": self.return_path or "",
        }

    def get_kwargs(self):
        return self.kwargs

    @classmethod
    def from_dict(cls, metadata: Dict[str, str], kwargs: Dict[str, Any], **extra) -> "RpcMessage":
        return cls(**metadata, **extra, kwargs=kwargs)


class ResultMessage(Message):
    """Representation of a Lightbus RPC Result message"""

    required_metadata = ["id", "rpc_message_id"]

    def __init__(
        self,
        *,
        result,
        api_name: str,
        procedure_name: str,
        rpc_message_id: str,
        id: str = "",
        error: bool = False,
        trace: str = None,
        native_id: str = None,
    ):
        super().__init__(id, native_id)
        self.api_name = api_name
        self.procedure_name = procedure_name
        self.rpc_message_id = rpc_message_id

        if isinstance(result, BaseException):
            self.result = repr(result)
            self.error = True
            self.trace = "".join(
                traceback.format_exception(
                    etype=type(result), value=result, tb=result.__traceback__
                )
            )
        else:
            self.result = result
            self.error = error
            self.trace = trace

    def __repr__(self):
        if self.error:
            return "<{} (ERROR): {}>".format(self.__class__.__name__, self.result)
        else:
            return "<{} (SUCCESS): {}>".format(self.__class__.__name__, self.result)

    def __str__(self):
        return str(self.result)

    def get_metadata(self) -> dict:
        metadata = {"id": self.id, "rpc_message_id": self.rpc_message_id, "error": self.error}
        if self.error:
            metadata["trace"] = self.trace
        return metadata

    def get_kwargs(self):
        return {"result": self.result}

    @classmethod
    def from_dict(
        cls, metadata: Dict[str, str], kwargs: Dict[str, Any], **extra
    ) -> "ResultMessage":
        return cls(**metadata, **extra, result=kwargs.get("result"))


class EventMessage(Message):
    """Representation of a Lightbus Event message"""

    required_metadata = ["id", "api_name", "event_name", "version"]

    def __init__(
        self,
        *,
        api_name: str,
        event_name: str,
        kwargs: Optional[dict] = None,
        version: int = 1,
        id: str = "",
        native_id: str = None,
    ):
        super().__init__(id, native_id)
        self.api_name = api_name
        self.event_name = event_name
        self.version = int(version)
        self.kwargs = kwargs or {}

    def __repr__(self):
        return "<{}: {}>".format(self.__class__.__name__, self)

    def __str__(self):
        return "{}({})".format(
            self.canonical_name, ", ".join("{}={}".format(k, v) for k, v in self.kwargs.items())
        )

    @property
    def canonical_name(self):
        return "{}.{}".format(self.api_name, self.event_name)

    def get_metadata(self) -> dict:
        return {
            "id": self.id,
            "api_name": self.api_name,
            "event_name": self.event_name,
            "version": self.version,
        }

    def get_kwargs(self):
        return self.kwargs

    @classmethod
    def from_dict(cls, metadata: Dict[str, str], kwargs: Dict[str, Any], **extra) -> "EventMessage":
        return cls(**metadata, **extra, kwargs=kwargs)



import asyncio
import inspect
import logging
from typing import List, Tuple, Callable, NamedTuple

from lightbus.schema.schema import Parameter
from lightbus.message import EventMessage
from lightbus.client.subclients.base import BaseSubClient
from lightbus.client.utilities import validate_event_or_rpc_name, queue_exception_checker, OnError
from lightbus.client.validator import validate_outgoing, validate_incoming
from lightbus.exceptions import (
    UnknownApi,
    EventNotFound,
    InvalidEventArguments,
    InvalidEventListener,
    ListenersAlreadyStarted,
    DuplicateListenerName,
)
from lightbus.log import L, Bold
from lightbus.client.commands import (
    SendEventCommand,
    AcknowledgeEventCommand,
    ConsumeEventsCommand,
    CloseCommand,
)
from lightbus.utilities.async_tools import run_user_provided_callable, cancel_and_log_exceptions
from lightbus.utilities.internal_queue import InternalQueue
from lightbus.utilities.casting import cast_to_signature
from lightbus.utilities.deforming import deform_to_bus
from lightbus.utilities.singledispatch import singledispatchmethod

logger = logging.getLogger(__name__)


class EventClient(BaseSubClient):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._event_listeners: List[Listener] = []
        self._event_listener_tasks = set()
        self._listeners_started = False

    async def fire_event(self, api_name, name, kwargs: dict = None, options: dict = None):
        kwargs = kwargs or {}
        try:
            api = self.api_registry.get(api_name)
        except UnknownApi:
            raise UnknownApi(
                "Lightbus tried to fire the event {api_name}.{name}, but no API named {api_name} was found in the "
                "registry. An API being in the registry implies you are an authority on that API. Therefore, "
                "Lightbus requires the API to be in the registry as it is a bad idea to fire "
                "events on behalf of remote APIs. However, this could also be caused by a typo in the "
                "API name or event name, or be because the API class has not been "
                "registered using bus.client.register_api(). ".format(**locals())
            )

        validate_event_or_rpc_name(api_name, "event", name)

        try:
            event = api.get_event(name)
        except EventNotFound:
            raise EventNotFound(
                "Lightbus tried to fire the event {api_name}.{name}, but the API {api_name} does not "
                "seem to contain an event named {name}. You may need to define the event, you "
                "may also be using the incorrect API. Also check for typos.".format(**locals())
            )

        parameter_names = {p.name if isinstance(p, Parameter) else p for p in event.parameters}

        if set(kwargs.keys()) != parameter_names:
            raise InvalidEventArguments(
                "Invalid event arguments supplied when firing event. Attempted to fire event with "
                "{} arguments: {}. Event expected {}: {}".format(
                    len(kwargs),
                    sorted(kwargs.keys()),
                    len(event.parameters),
                    sorted(parameter_names),
                )
            )

        kwargs = deform_to_bus(kwargs)
        event_message = EventMessage(
            api_name=api.meta.name, event_name=name, kwargs=kwargs, version=api.meta.version
        )

        validate_outgoing(self.config, self.schema, event_message)

        await self.hook_registry.execute("before_event_sent", event_message=event_message)
        logger.info(L("📤  Sending event {}.{}".format(Bold(api_name), Bold(name))))

        await self.producer.send(SendEventCommand(message=event_message, options=options)).wait()

        await self.hook_registry.execute("after_event_sent", event_message=event_message)

    def listen(
        self,
        events: List[Tuple[str, str]],
        listener: Callable,
        listener_name: str,
        options: dict = None,
        on_error: OnError = OnError.SHUTDOWN,
    ):
        if self._listeners_started:
            # We are actually technically able to support starting listeners after worker
            # startup, but it seems like it is a bad idea and a bit of an edge case.
            # We may revisit this if sufficient demand arises.
            raise ListenersAlreadyStarted(
                "You are trying to register a new listener after the worker has started running. "
                "Listeners should be setup in your @bus.client.on_start() hook, in your bus.py file."
            )

        sanity_check_listener(listener)

        for listener_api_name, _ in events:
            duplicate_listener = self.get_event_listener(listener_api_name, listener_name)
            if duplicate_listener:
                raise DuplicateListenerName(
                    f"A listener with name '{listener_name}' is already registered for API '{listener_api_name}'. "
                    f"You cannot have multiple listeners with the same name for a given API. Rename one of your "
                    f"listeners to resolve this problem."
                )

        for api_name, name in events:
            validate_event_or_rpc_name(api_name, "event", name)

        self._event_listeners.append(
            Listener(
                callable=listener,
                options=options or {},
                events=events,
                name=listener_name,
                on_error=on_error,
            )
        )

    def get_event_listener(self, api_name: str, listener_name: str):
        for listener in self._event_listeners:
            if listener.name == listener_name:
                for listener_api_name, _ in listener.events:
                    if listener_api_name == api_name:
                        return listener
        return None

    async def _on_message(
        self, event_message: EventMessage, listener: Callable, options: dict, on_error: OnError
    ):

        # TODO: Check events match those requested
        logger.info(
            L(
                "📩  Received event {}.{} with ID {}".format(
                    Bold(event_message.api_name), Bold(event_message.event_name), event_message.id
                )
            )
        )

        validate_incoming(self.config, self.schema, event_message)

        await self.hook_registry.execute("before_event_execution", event_message=event_message)

        if self.config.api(event_message.api_name).cast_values:
            parameters = cast_to_signature(parameters=event_message.kwargs, callable=listener)
        else:
            parameters = event_message.kwargs

        # Call the listener.
        # Pass the event message as a positional argument,
        # thereby allowing listeners to have flexibility in the argument names.
        # (And therefore allowing listeners to use the `event` parameter themselves)
        if on_error == OnError.SHUTDOWN:
            # Run the callback in the queue_exception_checker(). This will
            # put any errors into Lightbus' error queue, and therefore
            # cause a shutdown
            await queue_exception_checker(
                run_user_provided_callable(listener, args=[event_message], kwargs=parameters),
                self.error_queue,
                help=(
                    f"An error occurred while {listener} was handling an event. Lightbus will "
                    f"now shutdown. If you wish to continue you can use the on_error parameter "
                    f"when setting up your event. For example:\n\n"
                    f"    bus.my_api.my_event.listen(fn, listener_name='example', on_error=lightbus.OnError.ACKNOWLEDGE_AND_LOG)"
                ),
            )
        elif on_error == on_error.ACKNOWLEDGE_AND_LOG:
            try:
                await listener(event_message, **parameters)
            except asyncio.CancelledError:
                raise
            except Exception as e:
                # Log here. Acknowledgement will follow in below
                logger.exception(e)

        # Acknowledge the successfully processed message
        await self.producer.send(
            AcknowledgeEventCommand(message=event_message, options=options)
        ).wait()

        await self.hook_registry.execute("after_event_execution", event_message=event_message)

    async def close(self):
        await super().close()
        await cancel_and_log_exceptions(*self._event_listener_tasks)
        await self.producer.send(CloseCommand()).wait()

        await self.consumer.close()
        await self.producer.close()

    @singledispatchmethod
    async def handle(self, command):
        raise NotImplementedError(f"Did not recognise command {command.__class__.__name__}")

    async def start_registered_listeners(self):
        """Start all listeners which have been previously registered via listen()"""
        self._listeners_started = True
        for listener in self._event_listeners:
            await self._start_listener(listener)

    async def _start_listener(self, listener: "Listener"):
        # Setting the maxsize to 1 ensures the transport cannot load
        # messages faster than we can consume them
        queue: InternalQueue[EventMessage] = InternalQueue(maxsize=1)

        async def consume_events():
            while True:
                logger.debug("Event listener now waiting for event on the internal queue")
                event_message = await queue.get()
                logger.debug(
                    "Event listener has now received an event on the internal queue, processing now"
                )
                await self._on_message(
                    event_message=event_message,
                    listener=listener.callable,
                    options=listener.options,
                    on_error=listener.on_error,
                )
                queue.task_done()

        # Start the consume_events() consumer running
        task = asyncio.ensure_future(queue_exception_checker(consume_events(), self.error_queue))
        self._event_listener_tasks.add(task)

        await self.producer.send(
            ConsumeEventsCommand(
                events=listener.events,
                destination_queue=queue,
                listener_name=listener.name,
                options=listener.options,
            )
        ).wait()


class Listener(NamedTuple):
    callable: Callable
    options: dict
    events: List[Tuple[str, str]]
    name: str
    on_error: OnError


def sanity_check_listener(listener):
    if not callable(listener):
        raise InvalidEventListener(
            f"The specified event listener {listener} is not callable. Perhaps you called the function rather "
            f"than passing the function itself?"
        )

    total_positional_args = 0
    has_variable_positional_args = False  # Eg: *args
    for parameter in inspect.signature(listener).parameters.values():
        if parameter.kind in (
            inspect.Parameter.POSITIONAL_ONLY,
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
        ):
            total_positional_args += 1
        elif parameter.kind == inspect.Parameter.VAR_POSITIONAL:
            has_variable_positional_args = True

    if has_variable_positional_args:
        return

    if not total_positional_args:
        raise InvalidEventListener(
            f"The specified event listener {listener} must take at one positional argument. "
            f"This will be the event message. For example: "
            f"my_listener(event, other, ...)"
        )



"""Plugin to broadcast Lightbus' state on the internal.state API"""
import asyncio
import logging
import resource
import socket
from argparse import ArgumentParser, _ArgumentGroup, Namespace
from datetime import datetime

import os
from itertools import chain

from typing import TYPE_CHECKING

from lightbus.message import EventMessage
from lightbus.plugins import LightbusPlugin
from lightbus.plugins.metrics import MetricsPlugin
from lightbus.utilities.async_tools import cancel

if TYPE_CHECKING:
    # pylint: disable=unused-import,cyclic-import
    from lightbus import BusClient
    from lightbus.config import Config

logger = logging.getLogger(__name__)


class StatePlugin(LightbusPlugin):
    """Fire events to the bus regarding the state of this Lightbus worker

    This plugin allows the state of each Lightbus worker to be monitored
    via the bus itself. This is provided by the `internal.state` API.

    This plugin provides coarse monitoring in the following form:

      - Worker started events
      - Worker ping events - indicate the worker is alive. Sent every 60 seconds by default
      - Worker shutdown events

    See `lightbus.internal_apis.LightbusStateApi` for more details on the events provided.

    Note that this plugin provides the basic events required in order for the
    Lightbus admin interface to function. It is therefore suggested that you leave
    it in operation as the traffic generated should be minimal.

    Per-message events are available via the MetricsPlugin, which is substantially higher volume.
    """

    priority = 100

    def __init__(
        self,
        service_name: str,
        process_name: str,
        ping_enabled: bool = True,
        ping_interval: int = 60,
    ):
        self.service_name = service_name
        self.process_name = process_name
        self.ping_enabled = ping_enabled
        self.ping_interval = ping_interval

    @classmethod
    def from_config(cls, config: "Config", ping_enabled: bool = True, ping_interval: int = 60):
        return cls(
            service_name=config.service_name,
            process_name=config.process_name,
            ping_enabled=ping_enabled,
            ping_interval=ping_interval,
        )

    async def before_parse_args(self, *, parser: ArgumentParser, subparsers: _ArgumentGroup):
        """Add some plugin-related args so behaviour can be customised"""
        run_command_parser = subparsers.choices["run"]
        state_run_group = run_command_parser.add_argument_group(title="State plugin arguments")
        state_run_group.add_argument(
            "--ping-interval",
            help="Interval between worker ping events in seconds. Ping events alert the bus "
            "that this Lightbus worker is alive, and are used to update the lightbus admin interface.",
            metavar="SECONDS",
            type=int,
            default=self.ping_interval,
        )
        state_run_group.add_argument(
            "--no-ping",
            help="Disable sending ping events on the internal.state API. This "
            "may result in your worker not appearing in the lightbus admin interface, "
            "but will reduce traffic and log volume.",
            action="store_true",
        )

    async def receive_args(self, args: Namespace):
        if args.subcommand == "run":
            self.ping_enabled = not args.no_ping
            self.ping_interval = args.ping_interval or self.ping_interval

    async def before_worker_start(self, *, client: "BusClient"):
        event_transport = client.transport_registry.get_event_transport("internal.metrics")
        await event_transport.send_event(
            EventMessage(
                api_name="internal.state",
                event_name="worker_started",
                kwargs=self.get_state_kwargs(client),
            ),
            options={},
        )
        if self.ping_enabled:
            logger.info("Ping messages will be sent every {} seconds".format(self.ping_interval))
            self._ping_task = asyncio.ensure_future(self._send_ping(client))
        else:
            logger.warning(
                "Ping events have been disabled. This will reduce log volume and bus traffic, but "
                "may result in this Lightbus worker not appearing in the Lightbus admin interface."
            )

    async def after_worker_stopped(self, *, client: "BusClient"):
        event_transport = client.transport_registry.get_event_transport("internal.metrics")
        await event_transport.send_event(
            EventMessage(
                api_name="internal.state",
                event_name="worker_stopped",
                kwargs=dict(process_name=self.process_name, service_name=self.service_name),
            ),
            options={},
        )
        await cancel(self._ping_task)

    async def _send_ping(self, client: "BusClient"):
        event_transport = client.transport_registry.get_event_transport("internal.metrics")
        while True:
            await asyncio.sleep(self.ping_interval)
            await event_transport.send_event(
                EventMessage(
                    api_name="internal.state",
                    event_name="worker_ping",
                    kwargs=self.get_state_kwargs(client),
                ),
                options={},
            )

    def get_state_kwargs(self, client: "BusClient"):
        """Get the kwargs for a worker_started or ping message"""
        max_memory_use = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        event_listeners = chain(
            *[event_listener.events for event_listener in client.event_client._event_listeners]
        )
        return dict(
            process_name=self.process_name,
            service_name=self.service_name,
            metrics_enabled=client.plugin_registry.is_plugin_loaded(MetricsPlugin),
            api_names=[api.meta.name for api in client.api_registry.public()],
            listening_for=[
                "{}.{}".format(api_name, event_name) for api_name, event_name in event_listeners
            ],
            timestamp=datetime.utcnow().timestamp(),
            ping_enabled=self.ping_enabled,
            ping_interval=self.ping_interval,
            hostname=socket.gethostname(),
            pid=os.getpid(),
            max_memory_use=max_memory_use,
        )



"""


What are we actually trying to glue together here? We're trying to glue
two things together:

* The developer-facing API
* The transport layer

All interactions from user-provided callables will come via the developer-facing API.
We can therefore treat the handling of user-provided callables as a separate system.

What does this flow look like then? For example, if firing an event:

* Developer fires an event
* Client-API places it into an internal queue (optionally along with a asyncio.Condition)
* A consumer picks it up
* The consumer routes it to the correct transport

Notes:

* If we ensure the worker thread actually doesn't block then we will only need one worker thread.
  Currently it blocks, this is bad. We should aim to move the blocking into the
  top level API and not the worker thread.

"""

import asyncio
import logging
from typing import Optional

from lightbus.client.utilities import queue_exception_checker, ErrorQueueType
from lightbus.utilities.async_tools import cancel
from lightbus.utilities.internal_queue import InternalQueue

logger = logging.getLogger(__name__)


# Was invoker
class InternalProducer:
    """Base invoker class. Puts commands onto a queue

    Note that commands are execute in parallel. If you wish to know when a command has
    been executed you should await the command.on_done event.
    """

    # Warnings will be displayed if a queue grows to be equal to or greater than this size
    size_warning = 5

    # How often should the queue sizes be monitored
    monitor_interval = 0.1

    def __init__(self, queue: InternalQueue, error_queue: ErrorQueueType):
        """Initialise the invoker

        The callable specified by `on_exception` will be called with a single positional argument,
        which is the exception which occurred. This should take care of shutting down the invoker,
        as well as any other cleanup which needs to happen.
        """
        self._queue_monitor_task: Optional[asyncio.Task] = None
        self._monitor_ready = asyncio.Event()
        self.queue = queue
        self.error_queue = error_queue

    def start(self):
        """Starts the queue monitor"""
        # fmt: off
        self._queue_monitor_task = asyncio.ensure_future(queue_exception_checker(
            self._queue_monitor(),
            self.error_queue,
        ))
        # fmt: on

    async def close(self):
        if self._queue_monitor_task:
            await cancel(self._queue_monitor_task)
            self._queue_monitor_task = None
            self._monitor_ready = asyncio.Event()

    async def wait_until_ready(self):
        """Wait until this producer is ready to start receiving & handling commands"""
        await self._monitor_ready.wait()

    async def _queue_monitor(self):
        """Watches queues for growth and reports errors"""
        self._monitor_ready.set()

        previous_size = None
        while True:
            current_size = self.queue.qsize()

            show_size_warning = current_size >= self.size_warning and current_size != previous_size
            queue_has_shrunk = (
                previous_size is not None
                and current_size < previous_size
                and previous_size >= self.size_warning
            )

            if show_size_warning or queue_has_shrunk:
                if queue_has_shrunk:
                    if not show_size_warning:
                        everything_ok = " Queue is now at an OK size again."
                    else:
                        everything_ok = ""

                    logger.warning(
                        "Queue in %s has shrunk back down to %s commands.%s",
                        self.__class__.__name__,
                        current_size,
                        everything_ok,
                    )
                elif show_size_warning:
                    logger.warning(
                        "Queue in %s now has %s commands.", self.__class__.__name__, current_size
                    )

            previous_size = current_size
            await asyncio.sleep(self.monitor_interval)

    def send(self, command) -> asyncio.Event:
        logger.debug(f"Sending command {command}")
        event = asyncio.Event()
        self.queue.put_nowait((command, event))
        return event



import logging

logger = logging.getLogger(__name__)


def make_file_safe_api_name(api_name):
    """Make an api name safe for use in a file name"""
    return "".join([c for c in api_name if c.isalpha() or c.isdigit() or c in (".", "_", "-")])



import asyncio
import logging
import time
from typing import Mapping, Sequence, TYPE_CHECKING

from aioredis import PipelineError, ConnectionClosedError, ReplyError
from aioredis.util import decode

from lightbus.transports.base import RpcTransport, RpcMessage, Api
from lightbus.exceptions import TransportIsClosed
from lightbus.log import LBullets, L, Bold
from lightbus.serializers import BlobMessageSerializer, BlobMessageDeserializer
from lightbus.transports.redis.utilities import (
    RedisTransportMixin,
    retry_on_redis_connection_failure,
)
from lightbus.utilities.frozendict import frozendict
from lightbus.utilities.human import human_time
from lightbus.utilities.importing import import_from_string

if TYPE_CHECKING:
    # pylint: disable=unused-import,cyclic-import
    from lightbus.config import Config
    from lightbus.client import BusClient

logger = logging.getLogger("lightbus.transports.redis")


class RedisRpcTransport(RedisTransportMixin, RpcTransport):
    """ Redis RPC transport providing at-most-once delivery

    This transport uses a redis list and a blocking pop operation
    to distribute an RPC call to a single RPC consumer.

    Each call also has a corresponding expiry key created. Once the
    key expires it should be assumed that the RPC call has timed
    out and that therefore is should be discarded rather than
    be processed.
    """

    def __init__(
        self,
        *,
        redis_pool=None,
        url=None,
        serializer=BlobMessageSerializer(),
        deserializer=BlobMessageDeserializer(RpcMessage),
        connection_parameters: Mapping = frozendict(maxsize=100),
        batch_size=10,
        rpc_timeout=5,
        rpc_retry_delay=1,
        consumption_restart_delay=5,
    ):
        self.set_redis_pool(redis_pool, url, connection_parameters)
        self._latest_ids = {}
        self.serializer = serializer
        self.deserializer = deserializer
        self.batch_size = batch_size
        self.rpc_timeout = rpc_timeout
        self.rpc_retry_delay = rpc_retry_delay
        self.consumption_restart_delay = consumption_restart_delay
        super().__init__()

    @classmethod
    def from_config(
        cls,
        config: "Config",
        url: str = "redis://127.0.0.1:6379/0",
        connection_parameters: Mapping = frozendict(maxsize=100),
        batch_size: int = 10,
        serializer: str = "lightbus.serializers.BlobMessageSerializer",
        deserializer: str = "lightbus.serializers.BlobMessageDeserializer",
        rpc_timeout: int = 5,
        rpc_retry_delay: int = 1,
        consumption_restart_delay: int = 5,
    ):
        serializer = import_from_string(serializer)()
        deserializer = import_from_string(deserializer)(RpcMessage)

        return cls(
            url=url,
            serializer=serializer,
            deserializer=deserializer,
            connection_parameters=connection_parameters,
            batch_size=batch_size,
            rpc_timeout=rpc_timeout,
            consumption_restart_delay=consumption_restart_delay,
        )

    async def call_rpc(self, rpc_message: RpcMessage, options: dict):
        """Emit a call to a remote procedure

        This only sends the request, it does not await any result (see RedisResultTransport)
        """
        queue_key = f"{rpc_message.api_name}:rpc_queue"
        expiry_key = f"rpc_expiry_key:{rpc_message.id}"
        logger.debug(
            LBullets(
                L("Enqueuing message {} in Redis list {}", Bold(rpc_message), Bold(queue_key)),
                items=dict(**rpc_message.get_metadata(), kwargs=rpc_message.get_kwargs()),
            )
        )

        start_time = time.time()
        for try_number in range(3, 0, -1):
            last_try = try_number == 1
            try:
                await self._call_rpc(rpc_message, queue_key, expiry_key)
                break
            except (PipelineError, ConnectionClosedError, ConnectionResetError) as e:
                if not last_try:
                    logger.debug(
                        f"Retrying sending of message. Will retry {try_number} more times. "
                        f"Error was {type(e).__name__}: {e}"
                    )
                    await asyncio.sleep(self.rpc_retry_delay)
                else:
                    raise

        logger.debug(
            L(
                "Enqueued message {} in Redis in {} stream {}",
                Bold(rpc_message),
                human_time(time.time() - start_time),
                Bold(queue_key),
            )
        )

    async def _call_rpc(self, rpc_message: RpcMessage, queue_key, expiry_key):
        with await self.connection_manager() as redis:
            p = redis.pipeline()
            p.rpush(key=queue_key, value=self.serializer(rpc_message))
            p.set(expiry_key, 1)
            # TODO: Ditch this somehow. We kind of need the ApiConfig here, but a transports can be
            #       used for multiple APIs. So we could pass it in for each use of the transports,
            #       but even then it doesn't work for the event listening. This is because an event listener
            #       can span multiple APIs
            p.expire(expiry_key, timeout=self.rpc_timeout)
            await p.execute()

    async def consume_rpcs(self, apis: Sequence[Api]) -> Sequence[RpcMessage]:
        """Consume RPCs for the given APIs"""
        if self._closed:
            # Triggered during shutdown
            raise TransportIsClosed("Transport is closed. Cannot consume RPCs")

        return await retry_on_redis_connection_failure(
            fn=self._consume_rpcs,
            args=[apis],
            retry_delay=self.consumption_restart_delay,
            action="consuming RPCs",
        )

    async def _consume_rpcs(self, apis: Sequence[Api]) -> Sequence[RpcMessage]:
        # Get the name of each list queue
        queue_keys = ["{}:rpc_queue".format(api.meta.name) for api in apis]

        logger.debug(
            LBullets(
                "Consuming RPCs from",
                items=["{} ({})".format(s, self._latest_ids.get(s, "$")) for s in queue_keys],
            )
        )

        with await self.connection_manager() as redis:
            try:
                try:
                    stream, data = await redis.blpop(*queue_keys)
                except RuntimeError:
                    # For some reason aio-redis likes to eat the CancelledError and
                    # turn it into a Runtime error:
                    # https://github.com/aio-libs/aioredis/blob/9f5964/aioredis/connection.py#L184
                    raise asyncio.CancelledError(
                        "aio-redis task was cancelled and decided it should be a RuntimeError"
                    )
            except asyncio.CancelledError:
                # We need to manually close the connection here otherwise the aioredis
                # pool will emit warnings saying that this connection still has pending
                # commands (i.e. the above blocking pop)
                redis.close()
                raise

            stream = decode(stream, "utf8")
            rpc_message = self.deserializer(data)
            expiry_key = f"rpc_expiry_key:{rpc_message.id}"
            key_deleted = await redis.delete(expiry_key)

            if not key_deleted:
                return []

            logger.debug(
                LBullets(
                    L("⬅ Received RPC message on stream {}", Bold(stream)),
                    items=dict(**rpc_message.get_metadata(), kwargs=rpc_message.get_kwargs()),
                )
            )

            return [rpc_message]



import asyncio
import logging
from typing import Optional, Callable

from lightbus.client.utilities import queue_exception_checker, ErrorQueueType
from lightbus.utilities.async_tools import cancel
from lightbus.utilities.internal_queue import InternalQueue

logger = logging.getLogger(__name__)


# Was handler
class InternalConsumer:
    """Handle commands coming from the transport invoker

    These are commands which are being sent by the client for
    consumption by the transport
    """

    def __init__(self, queue: InternalQueue, error_queue: ErrorQueueType):
        self._consumer_task: Optional[asyncio.Task] = None
        self._running_commands = set()
        self._ready = asyncio.Event()
        self.queue = queue
        self.error_queue = error_queue

    def start(self, handler: Callable):
        """Set the handler function and start the invoker

        Use `stop()` to shutdown the invoker.
        """
        logger.debug(
            f"Starting consumer for handler {handler.__qualname__}(). This should report ready"
            " shortly..."
        )
        self._consumer_task = asyncio.ensure_future(
            queue_exception_checker(self._consumer_loop(self.queue, handler), self.error_queue)
        )
        self._running_commands = set()

    async def close(self):
        """Shutdown the invoker and cancel any currently running tasks

        The shutdown procedure will stop any new tasks being created
        then shutdown all existing tasks
        """
        # Stop consuming commands from the queue
        # (this will also stop *new* tasks being created)
        if self._consumer_task is not None:
            await cancel(self._consumer_task)
            self._consumer_task = None
            self._ready = asyncio.Event()

        # Now we have stopped consuming commands we can
        # cancel any running tasks safe in the knowledge that
        # no new tasks will get created
        await cancel(*self._running_commands)

    async def _consumer_loop(self, queue, handler):
        """Continually fetch commands from the queue and handle them"""
        logger.debug(f"Consumer loop is ready with handler {handler.__qualname__}()")
        self._ready.set()

        while True:
            on_done: asyncio.Event
            command, on_done = await queue.get()
            self.handle_in_background(queue, handler, command, on_done)

    async def wait_until_ready(self):
        """Wait until this consumer is ready to start receiving & handling commands"""
        await self._ready.wait()

    def handle_in_background(self, queue: InternalQueue, handler, command, on_done: asyncio.Event):
        """Handle a received command by calling the provided handler

        This execution happens in the background.
        """
        logger.debug(f"Handling command {command}")

        def when_task_finished(fut: asyncio.Future):
            self._running_commands.remove(fut)
            try:
                # Retrieve any error which may have occurred.
                # We ignore the error because we assume any exceptions which the
                # handler threw will have already been placed into the error queue
                # by the queue_exception_checker().
                # Regardless, we must retrieve the result in order to keep Python happy.
                fut.result()
            except:
                pass

            # We use call_soon_threadsafe() to ensure we call the Event's set()
            # in a threadsafe fashion. This is because the Event object may have
            # been created in another thread and be attached to another event loop
            on_done._loop.call_soon_threadsafe(on_done.set)

        # fmt: off
        background_call_task = asyncio.ensure_future(queue_exception_checker(
            handler(command),
            self.error_queue,
        ))
        # fmt: on
        background_call_task.add_done_callback(when_task_finished)
        self._running_commands.add(background_call_task)



from functools import wraps


def uses_django_db(f):
    """Ensures Django discards any broken database connections

    Django normally cleans up connections once a web request has
    been processed. However, here we are not serving web requests
    and are outside of Django's request handling logic. We therefore
    need to make sure we cleanup any broken database connections.
    """
    # TODO: Move this into middleware
    #       (Tracked in: https://github.com/adamcharnock/lightbus/issues/6)

    # Import Django locally as it is not a dependency of Lightbus.
    # This will only get run on startup anyway, and we will assume that
    # if someone uses this decorator then they have Django installed.
    from django import db

    @wraps(f)
    def wrapped(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        finally:
            db.connections.close_all()

    return wrapped



from lightbus.transports.redis.event import RedisEventTransport
from lightbus.transports.redis.result import RedisResultTransport
from lightbus.transports.redis.rpc import RedisRpcTransport
from lightbus.transports.redis.schema import RedisSchemaTransport
from lightbus.transports.redis.utilities import RedisEventMessage



import argparse
import json
import logging
import re
from asyncio import sleep
from hashlib import sha1

import sys
from fnmatch import fnmatch
from pathlib import Path
from typing import Optional, List, get_type_hints

from lightbus import EventMessage, BusPath, EventTransport, ValidationError
from lightbus.commands import utilities as command_utilities
from lightbus.plugins import PluginRegistry
from lightbus.utilities.async_tools import block
from lightbus.utilities.casting import cast_to_signature
from lightbus_vendored.jsonpath import jsonpath


CACHE_PATH = Path("~/.lightbus/inspect_cache").expanduser()

logger = logging.getLogger(__name__)


class Command:
    def setup(self, parser, subparsers):
        parser_inspect = subparsers.add_parser(
            "inspect",
            help="Inspect events on the bus",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        group = parser_inspect.add_argument_group(title="Inspect command arguments")
        group.add_argument(
            "--json-path",
            "-j",
            help=("Search event body json for the givn value. Eg. address.city=London"),
            metavar="JSON_SEARCH",
        )
        group.add_argument(
            "--id",
            "-i",
            help=("Find a single event with this Lightbus event ID"),
            metavar="LIGHTBUS_EVENT_ID",
        )
        group.add_argument(
            "--native-id",
            "-n",
            help=("Find a single event with this broker-native ID"),
            metavar="NATIVE_EVENT_ID",
        )
        group.add_argument(
            "--api",
            "-a",
            help=("Find events for this API name. Supports the '*' wildcard."),
            metavar="API_NAME",
        )
        group.add_argument(
            "--event",
            "-e",
            help=("Find events for this event name. Supports the '*' wildcard."),
            metavar="EVENT_NAME",
        )
        group.add_argument(
            "--version",
            "-v",
            help=("Find events with the specified version number. Can be prefixed by <, >, <=, >="),
            metavar="VERSION_NUMBER",
            type=int,
        )
        group.add_argument(
            "--format",
            "-F",
            help="Formatting style. One of json, pretty, or human.",
            metavar="FORMAT",
            default="json",
            choices=("json", "pretty", "human"),
        )
        group.add_argument(
            "--cache-only", "-c", help=("Search the local cache only"), action="store_true"
        )
        group.add_argument(
            "--follow",
            "-f",
            help=(
                "Continually listen for new events matching the search criteria. "
                "May only be used on a single API"
            ),
            action="store_true",
        )
        group.add_argument(
            "--validate",
            "-d",
            help=(
                "Validate displayed events against the bus schema. Only available "
                "when --format is set to 'human' (experimental)"
            ),
            action="store_true",
        )
        group.add_argument(
            "--show-casting",
            "-C",
            help=(
                "Show how the message values will be casted for each event listener (experimental)"
            ),
            action="store_true",
        )
        group.add_argument("--internal", "-I", help="Include internal APIs", action="store_true")

        command_utilities.setup_common_arguments(parser_inspect)
        parser_inspect.set_defaults(func=self.handle)

    def handle(self, args, config, plugin_registry: PluginRegistry):
        """Entrypoint for the inspect command"""
        command_utilities.setup_logging(args.log_level or "warning", config)
        bus_module, bus = command_utilities.import_bus(args)
        api_names: List[str]

        block(bus.client.lazy_load_now())
        block(bus.client.hook_registry.execute("before_worker_start"))

        # Locally registered APIs
        api_names = [api.meta.name for api in bus.client.api_registry.all()]

        # APIs registered to other services on the bus
        for api_name in bus.client.schema.api_names:
            if api_name not in api_names:
                api_names.append(api_name)

        if not args.internal and not args.api:
            # Hide internal APIs if we don't want them
            api_names = [api_name for api_name in api_names if not api_name.startswith("internal.")]

        if args.api and args.api not in api_names:
            sys.stderr.write(
                f"Specified API was not found locally or within the schema on the bus.\n"
                f"Ensure a Lightbus worker is running for this API.\n"
                f"Cannot continue.\n"
            )
            sys.exit(1)

        api_names_to_inspect = []
        for api_name in api_names:
            if not args.api or self.wildcard_match(args.api, api_name):
                api_names_to_inspect.append(api_name)

        if len(api_names_to_inspect) != 1 and args.follow:
            sys.stderr.write(
                f"The --follow option is only available when following a single API.\n"
                f"Please specify the --api option to select a single API to follow.\n"
            )
            sys.exit(1)

        try:
            for api_name in api_names_to_inspect:
                if not args.api or self.wildcard_match(args.api, api_name):
                    logger.debug(f"Inspecting {api_name}")
                    block(self.search_in_api(args, api_name, bus))
                else:
                    logger.debug(f"API {api_name} did not match {args.api}. Skipping")
        except KeyboardInterrupt:
            logger.info("Stopped by user")

    async def search_in_api(self, args, api_name: str, bus: BusPath):
        transport_pool = bus.client.transport_registry.get_event_transport(api_name)
        async with transport_pool as transport:
            async for message in self.get_messages(args, api_name, args.event, transport, bus):
                if self.match_message(args, message):
                    self.output(args, transport, message, bus)

    async def get_messages(
        self,
        args,
        api_name: str,
        event_name: Optional[str],
        transport: EventTransport,
        bus: BusPath,
    ):
        """Yields messages from various sources

        Messages are returned from sources in this order:

         - Any on-disk cache
         - Reading history from the event transport

        """
        CACHE_PATH.mkdir(parents=True, exist_ok=True)
        event_name = event_name or "*"

        # Construct the cache file name
        file_name_hash = sha1((api_name + "\0" + event_name).encode("utf8")).hexdigest()[:8]
        file_name_api = re.sub("[^a-zA-Z0-9_]", "_", api_name)
        file_name_event = event_name.replace("*", "all")

        cache_file_name = f"{file_name_hash}-{file_name_api}-{file_name_event}.json"
        cache_file = CACHE_PATH / cache_file_name

        logger.debug(f"Loading from cache file {cache_file}. Exists: {cache_file.exists()}")

        # Sanity check
        if not cache_file.exists() and args.cache_only:
            sys.stderr.write(
                f"No cache file exists for {api_name}.{event_name}, but --cache-only was specified\n"
            )
            sys.exit(1)

        def _progress(force=False):
            if force or (cache_yield_count + transport_yield_count) % 1000 == 0:
                logger.debug(
                    f"Yielded {cache_yield_count} from cache and {transport_yield_count} from bus"
                )

        # Start by reading from cache
        cache_yield_count = 0
        transport_yield_count = 0
        start = None
        if cache_file.exists():
            with cache_file.open() as f:
                for line in f:
                    event_message = transport.deserializer(json.loads(line))

                    if not args.cache_only:
                        if not hasattr(event_message, "datetime"):
                            # Messages do not provide a datetime, stop loading from the cache as
                            # this is required
                            logger.warning(
                                "Event transport does not provide message datetimes. Will not load from cache."
                            )
                            break
                        start = (
                            max(event_message.datetime, start) if start else event_message.datetime
                        )

                    cache_yield_count += 1
                    _progress()
                    yield event_message

        if args.cache_only:
            return

        def _write_to_cache(f, event_message):
            f.write(json.dumps(transport.serializer(event_message)))
            f.write("\n")
            f.flush()

        # Now get messages from the transport, writing to the cache as we go
        allow_following = True

        if start:
            logger.debug(
                f"Finished reading from cache. Now reading from {start} on {api_name}.{event_name}"
            )

        while True:
            with cache_file.open("a") as f:
                async for event_message in transport.history(
                    api_name=api_name, event_name=event_name, start=start, start_inclusive=False
                ):
                    _write_to_cache(f, event_message)
                    transport_yield_count += 1
                    _progress()
                    yield event_message
                    if hasattr(event_message, "datetime"):
                        start = event_message.datetime
                    else:
                        # Following requires the datetime property on event messages
                        allow_following = False

            if args.follow:
                if allow_following:
                    # We want to keep waiting for new messages, so wait a second then do it again
                    await sleep(1)
                else:
                    logger.warning(
                        "Event transport does not provide message datetimes. Following not supported."
                    )
                    break
            else:
                # No waiting for new messages, so break out of the while loop
                break

        _progress(force=True)

    def wildcard_match(self, pattern: str, s: str) -> bool:
        return fnmatch(s, pattern)

    def version_match(self, pattern: str, version: int) -> bool:
        """Does the given version match the given pattern"""
        match = re.match(r"(<|>|<=|>=|!=|=)(\d+)", pattern)
        if not match:
            sys.stderr.write("Invalid version\n")
            sys.exit(1)

        comparator, version_ = match.groups()

        try:
            version_ = int(version_)
        except TypeError:
            sys.stderr.write("Invalid version\n")
            sys.exit(1)

        return self.compare(comparator, left_value=version, right_value=version_)

    def json_match(self, query: str, data: dict) -> bool:
        """Does the given data match the given jsonpath query"""
        match = re.match(r"(.+?)(<|>|<=|>=|!=|=)(.+)", query)
        if not match:
            sys.stderr.write("Invalid json query\n")
            sys.exit(1)

        query, comparator, value = match.groups()
        found_values = jsonpath(data, query)

        if not found_values:
            return False

        for found_value in found_values:
            if self.compare(comparator, left_value=found_value, right_value=value):
                return True

        return False

    def match_message(self, args, message: EventMessage) -> bool:
        """Does a message match the given search criteria?"""
        if args.id and args.id != message.id:
            return False

        if args.native_id and args.native_id != message.native_id:
            return False

        if args.event and not self.wildcard_match(args.event, message.event_name):
            return False

        if args.json_path and not self.json_match(args.json_path, message.kwargs):
            return False

        if args.version and not self.version_match(args.json, message.version):
            return False

        return True

    def output(self, args, transport: EventTransport, message: EventMessage, bus: BusPath):
        """Print out the given message"""

        serialized = transport.serializer(message)
        if args.format in ("json", "pretty"):
            if args.format == "pretty":
                dumped = json.dumps(serialized, indent=4)
            else:
                dumped = json.dumps(serialized)
            sys.stdout.write(dumped)
            sys.stdout.write("\n")
            sys.stdout.flush()

        elif args.format == "human":
            print(Colors.BGreen, end="")
            print(f" {message.api_name}.{message.event_name} ".center(80, "="))
            if hasattr(message, "datetime"):
                print(f" {message.datetime.strftime('%c')} ".center(80, " "))
            print(Colors.Reset, end="")

            print(f"\n{Colors.BWhite}Metadata:{Colors.Reset}")
            for k, v in message.get_metadata().items():
                print(f"    {str(k).ljust(20)}: {v}")

            print(f"\n{Colors.BWhite}Data:{Colors.Reset}")
            for k, v in message.get_kwargs().items():
                if isinstance(v, (dict, list)):
                    v = json.dumps(v, indent=4)
                    pad = " " * 24
                    v = "".join(pad + v for v in v.splitlines(keepends=True)).lstrip()
                print(f"    {str(k).ljust(20)}: {v}")

            if args.validate or args.show_casting:
                print(f"\n{Colors.BWhite}Extra:{Colors.Reset}")

            if args.validate:
                try:
                    bus.client.schema.validate_parameters(
                        message.api_name, message.event_name, message.kwargs
                    )
                except ValidationError as e:
                    validation_message = f"{Colors.Red}{e}{Colors.Reset}"
                else:
                    validation_message = f"{Colors.Green}Passed{Colors.Reset}"

                print(f"    Validation:         {validation_message}")

            if args.show_casting:
                for listener in bus.client.event_client._event_listeners:
                    if (message.api_name, message.event_name) not in listener.events:
                        continue

                    hints = get_type_hints(listener.callable)
                    casted = cast_to_signature(
                        parameters=message.kwargs, callable=listener.callable
                    )
                    print(
                        f"\n    {Colors.BWhite}Casting for listener: {listener.name}{Colors.Reset}"
                    )

                    for key, value in message.kwargs.items():
                        was = type(value)
                        via = hints[key]
                        now = type(casted[key])
                        color = Colors.Green if via == now else Colors.Red
                        print(
                            f"        "
                            f"{color}{str(key).ljust(20)}: "
                            f"Received a '{was.__name__}', "
                            f"casted to a '{via.__name__}', "
                            f"result was a '{now.__name__}'"
                            f"{Colors.Reset}"
                        )

            print("\n")

        else:
            sys.stderr.write(f"Unknown output format '{args.format}'\n")
            sys.exit(1)

    def compare(self, comparator: str, left_value, right_value):
        """Utility for performing arbitrary comparisons"""
        lookup = {
            "=": lambda: left_value == right_value,
            "!=": lambda: left_value != right_value,
            "<": lambda: left_value < right_value,
            ">": lambda: left_value > right_value,
            "<=": lambda: left_value <= right_value,
            ">=": lambda: left_value >= right_value,
        }

        if comparator not in lookup:
            sys.stderr.write(f"Unknown comparator '{comparator}'\n")
            sys.exit(1)

        return lookup[comparator]()


class Colors:
    # Reset
    Reset = "\033[0m"  # Text Reset

    # Regular Colors
    Black = "\033[0;30m"
    Red = "\033[0;31m"
    Green = "\033[0;32m"
    Yellow = "\033[0;33m"
    Blue = "\033[0;34m"
    Purple = "\033[0;35m"
    Cyan = "\033[0;36m"
    White = "\033[0;37m"

    # Bold
    BBlack = "\033[1;30m"
    BRed = "\033[1;31m"
    BGreen = "\033[1;32m"
    BYellow = "\033[1;33m"
    BBlue = "\033[1;34m"
    BPurple = "\033[1;35m"
    BCyan = "\033[1;36m"
    BWhite = "\033[1;37m"
